# generated by datamodel-codegen:
#   filename:  https://raw.githubusercontent.com/HumanBrainProject/openMINDS/3fa86f956b407b2debf47c2e1b6314e37579c707/v3/core/v4/actors/organization.schema.json

from typing import Any, Dict, List, Optional

from pydantic import Field
from siibra.openminds.base import SiibraBaseModel


class Model(SiibraBaseModel):
    id: str = Field(..., alias='@id', description='Metadata node identifier.')
    type: str = Field(..., alias='@type')
    digital_identifier: Optional[List[Any]] = Field(
        None,
        alias='digitalIdentifier',
        description='Digital handle to identify objects or legal persons.',
        min_items=1,
        title='digitalIdentifier',
    )
    full_name: str = Field(
        ...,
        alias='fullName',
        description='Whole, non-abbreviated name of something or somebody.',
        title='fullName',
    )
    has_parent: Optional[Dict[str, Any]] = Field(
        None,
        alias='hasParent',
        description='Reference to a parent object or legal person.',
        title='hasParent',
    )
    homepage: Optional[Dict[str, Any]] = Field(
        None,
        alias='homepage',
        description='Main website of something or someone.',
        title='homepage',
    )
    short_name: Optional[str] = Field(
        None,
        alias='shortName',
        description='Shortened or fully abbreviated name of something or somebody.',
        title='shortName',
    )



# generated by datamodel-codegen:
#   filename:  https://raw.githubusercontent.com/HumanBrainProject/openMINDS/3fa86f956b407b2debf47c2e1b6314e37579c707/v3/core/v4/products/metaDataModel.schema.json

from typing import Any, Dict, List, Optional

from pydantic import Field, constr
from siibra.openminds.base import SiibraBaseModel


class Model(SiibraBaseModel):
    id: str = Field(..., alias='@id', description='Metadata node identifier.')
    type: str = Field(..., alias='@type')
    custodian: Optional[List[Any]] = Field(
        None,
        alias='custodian',
        description="The 'custodian' is a legal person who is responsible for the content and quality of the data, metadata, and/or code of a research product.",
        min_items=1,
        title='custodian',
    )
    description: constr(max_length=2000) = Field(
        ...,
        alias='description',
        description='Longer statement or account giving the characteristics of someone or something.',
        title='description',
    )
    developer: List[Any] = Field(
        ...,
        alias='developer',
        description='Legal person that creates or improves products or services (e.g., software, applications, etc.).',
        min_items=1,
        title='developer',
    )
    digital_identifier: Optional[Dict[str, Any]] = Field(
        None,
        alias='digitalIdentifier',
        description='Digital handle to identify objects or legal persons.',
        title='digitalIdentifier',
    )
    full_name: str = Field(
        ...,
        alias='fullName',
        description='Whole, non-abbreviated name of something or somebody.',
        title='fullName',
    )
    has_version: List[Any] = Field(
        ...,
        alias='hasVersion',
        description='Reference to variants of an original.',
        min_items=1,
        title='hasVersion',
    )
    homepage: Optional[Dict[str, Any]] = Field(
        None,
        alias='homepage',
        description='Main website of something or someone.',
        title='homepage',
    )
    how_to_cite: Optional[str] = Field(
        None,
        alias='howToCite',
        description='Preferred format for citing a particular object or legal person.',
        title='howToCite',
    )
    short_name: constr(max_length=30) = Field(
        ...,
        alias='shortName',
        description='Shortened or fully abbreviated name of something or somebody.',
        title='shortName',
    )



# generated by datamodel-codegen:
#   filename:  https://raw.githubusercontent.com/HumanBrainProject/openMINDS/3fa86f956b407b2debf47c2e1b6314e37579c707/v3/SANDS/v3/miscellaneous/qualitativeRelationAssessment.schema.json

from typing import Any, Dict, Optional

from pydantic import Field
from siibra.openminds.base import SiibraBaseModel


class Model(SiibraBaseModel):
    id: str = Field(..., alias='@id', description='Metadata node identifier.')
    type: str = Field(..., alias='@type')
    criteria: Optional[Dict[str, Any]] = Field(
        None,
        alias='criteria',
        description='Aspects or standards on which a judgement or decision is based.',
        title='criteria',
    )
    in_relation_to: Dict[str, Any] = Field(
        ...,
        alias='inRelationTo',
        description='Reference to a related element.',
        title='inRelationTo',
    )
    qualitative_overlap: Dict[str, Any] = Field(
        ...,
        alias='qualitativeOverlap',
        description='Semantic characterization of how much two things occupy the same space.',
        title='qualitativeOverlap',
    )



"""
n.b.

this script requires datamodel-code-generator, which is not installed via requirements.
"""


from os import path
from pathlib import Path
import os
from collections import namedtuple
from typing import List
import requests
from datamodel_code_generator import generate, InputFileType
from datamodel_code_generator.format import PythonVersion
from siibra.openminds.base import SiibraBaseModel

path_to_currdir = path.dirname(
    path.abspath(__file__)
)
raw_url = 'https://raw.githubusercontent.com/HumanBrainProject/openMINDS/{ref}/{openminds_v}/{domain}/{schema}.schema.json'
"""
domain typically follows
[a-zA-Z]+/v[0-9]+

e.g. SANDS/v3


schema typeicall follows
[a-zA-Z]+/[a-zA-Z]+

e.g. miscellaneous/coordinatePoint
"""

OpenmindsSchema = namedtuple('OpenmindsSchema', ['domain', 'schema'])
openminds_ref = '3fa86f956b407b2debf47c2e1b6314e37579c707'
openminds_v = 'v3'

def process_schema(model: OpenmindsSchema):
    url = raw_url.format(
        ref=openminds_ref,
        openminds_v=openminds_v,
        domain=model.domain,
        schema=model.schema,
    )
    resp = requests.get(url)
    assert resp.status_code == 200
    resp_text = resp.text

    target_dir = f'{path_to_currdir}/{model.domain}/{path.dirname(model.schema)}'
    os.makedirs(target_dir, exist_ok=True)

    output_filename = f'{path_to_currdir}/{model.domain}/{model.schema}.py'
    generate(
        resp_text,
        target_python_version=PythonVersion.PY_36,
        input_file_type=InputFileType.JsonSchema,
        input_filename=url,
        snake_case_field=True,
        output=Path(output_filename),
        disable_timestamp=True,
        base_class="siibra.openminds.base.SiibraBaseModel"
    )
    with open(output_filename, "r") as fp:
        txt = fp.read()
    with open(output_filename, "w") as fp:
        fp.write(txt.replace("https://openminds.ebrains.eu/vocab/", ""))


def add_init():
    def process_dir(dirpath: str):

        # ignore directories starts with _
        if dirpath[:1] == "_":
            return

        # ignore if path is not directory
        if not os.path.isdir(dirpath):
            return
        
        # check if __init__.py exists. if not, create it.
        init_path = os.path.join(dirpath, "__init__.py")
        if not os.path.exists(init_path):
            with open(init_path, "w"):
                pass
        
        # iterate all existing items in directory
        for fname in os.listdir(dirpath):
            process_dir(path.join(dirpath, fname))
    
    for fname in os.listdir(path_to_currdir):
        process_dir(path.join(path_to_currdir, fname))
    

def main():
    resp = requests.get(f'https://api.github.com/repos/HumanBrainProject/openMINDS/git/trees/{openminds_ref}?recursive=1')
    assert resp.status_code == 200
    resp_json = resp.json()
    tree = resp_json.get('tree')
    import re
    starts_with = re.compile(f'^{openminds_v}')
    ends_with = re.compile(r'\.schema\.json$')
    filtered_tree: List[str] = [t.get('path') for t in tree
        if starts_with.search(t.get('path'))
        and ends_with.search(t.get('path'))
        and t.get('type') == 'blob'    
    ]
    find_domain_schema = re.compile(f'^{openminds_v}/([a-zA-Z]+\/v[0-9]+)\/(.+)\.schema\.json$')
    
    
    for filepath in filtered_tree:
        full_str = find_domain_schema.search(filepath)
        domain, schema = full_str.groups()
        
        process_schema(
            OpenmindsSchema(domain, schema)
        )
    add_init()
    

if __name__ == "__main__":
    main()


# generated by datamodel-codegen:
#   filename:  https://raw.githubusercontent.com/HumanBrainProject/openMINDS/3fa86f956b407b2debf47c2e1b6314e37579c707/v3/core/v4/miscellaneous/quantitativeValue.schema.json

from typing import Any, Dict, List, Optional

from pydantic import Field
from siibra.openminds.base import SiibraBaseModel


class Model(SiibraBaseModel):
    id: str = Field(..., alias='@id', description='Metadata node identifier.')
    type: str = Field(..., alias='@type')
    type_of_uncertainty: Optional[Dict[str, Any]] = Field(
        None,
        alias='typeOfUncertainty',
        description='Distinct technique used to quantify the uncertainty of a measurement.',
        title='typeOfUncertainty',
    )
    uncertainty: Optional[List[float]] = Field(
        None,
        alias='uncertainty',
        description='Quantitative value range defining the uncertainty of a measurement.',
        max_items=2,
        min_items=2,
        title='uncertainty',
    )
    unit: Optional[Dict[str, Any]] = Field(
        None,
        alias='unit',
        description='Determinate quantity adopted as a standard of measurement.',
        title='unit',
    )
    value: float = Field(
        ...,
        alias='value',
        description='Entry for a property.',
        title='value',
    )



# generated by datamodel-codegen:
#   filename:  https://raw.githubusercontent.com/HumanBrainProject/openMINDS/3fa86f956b407b2debf47c2e1b6314e37579c707/v3/core/v4/research/stimulation.schema.json

from typing import Any, Dict, Optional

from pydantic import Field
from siibra.openminds.base import SiibraBaseModel


class Model(SiibraBaseModel):
    id: str = Field(..., alias='@id', description='Metadata node identifier.')
    type: str = Field(..., alias='@type')
    description: Optional[str] = Field(
        None,
        alias='description',
        description='Longer statement or account giving the characteristics of someone or something.',
        title='description',
    )
    lookup_label: Optional[str] = Field(
        None,
        alias='lookupLabel',
        title='lookupLabel',
    )
    stimulation_approach: Dict[str, Any] = Field(
        ...,
        alias='stimulationApproach',
        title='stimulationApproach',
    )
    stimulus_type: Dict[str, Any] = Field(
        ...,
        alias='stimulusType',
        title='stimulusType',
    )



# Copyright 2018-2021
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .serializable_concept import JSONSerializable
from ..commons import logger
from ..retrieval import EbrainsKgQuery
from ..retrieval.requests import DECODERS, EbrainsRequest
from ..openminds.core.v4.products.datasetVersion import Model as DatasetVersionModel
from ..openminds.base import ConfigBaseModel

import hashlib
import re
from datetime import date
from typing import Any, Dict, List, Optional
from pydantic import Field

class Url(ConfigBaseModel):
    doi: str
    cite: Optional[str]

class Dataset(JSONSerializable):
    """Parent class for datasets. Each dataset has an identifier."""

    REGISTRY = {}

    def __init__(self, identifier, description=""):
        self.id = identifier
        self._description_cached = description

    def __str__(self):
        return f"{self.__class__.__name__}: {self.id}"

    def __init_subclass__(cls, type_id=None):
        if type_id in Dataset.REGISTRY:
            logger.warning(
                f"Type id '{type_id}' already provided by {Dataset.REGISTRY[type_id].__name__}, but {cls.__name__} suggests itself as well"
            )
        if type_id is not None:
            logger.debug(f"Registering specialist {cls.__name__} for type id {type_id}")
            Dataset.REGISTRY[type_id] = cls
        cls.type_id = type_id
        return super().__init_subclass__()

    @property
    def is_volume(self):
        """Return True if this dataset represents a brain volume source.

        This property is overwritten by siibra.volumes.VolumeSrc
        """
        return False

    @property
    def is_surface(self):
        """Return True if this dataset represents a brain volume surface. 
        
        This property is overwritten by siibra.volumes.VolumeSrc
        """
        return False

    @property
    def publications(self):
        """
        List of publications for this dataset.
        Empty list here, but implemented in some derived classes.
        """
        return []

    @property
    def urls(self):
        """
        List of URLs related to this dataset.
        Empty list here, but implemented in some derived classes.
        """
        return []

    @property
    def description(self):
        """
        Textual description of Dataset.
        Empty string here, but implemented in some derived classes.
        """
        return self._description_cached

    @classmethod
    def extract_type_id(cls, spec):
        for key in ["@type", "kgSchema"]:
            if key in spec:
                return spec[key]
        raise RuntimeError(f"No type defined in dataset specification: {spec}")

    @classmethod
    def get_model_type(Cls):
        return "https://openminds.ebrains.eu/core/DatasetVersion"

    @property
    def model_id(self):
        _id = hashlib.md5(
            str(
                self.id if self.id else f"{str(self)}{self.description}"
            ).encode("utf-8")
        ).hexdigest()
        return f'{self.get_model_type()}/{_id}'

    def to_model(self, **kwargs) -> 'DatasetJsonModel':
        metadata=DatasetVersionModel(
            id=self.model_id,
            type=self.get_model_type(),
            accessibility={ "@id": self.embargo_status[0].get("@id") } \
                if hasattr(self, 'embargo_status') \
                and self.embargo_status is not None \
                and len(self.embargo_status) == 1 \
                else { "@id": "https://openminds.ebrains.eu/instances/productAccessibility/freeAccess" },
            data_type=[{
                "@id": "https://openminds.ebrains.eu/instances/semanticDataType/derivedData"
            }],
            digital_identifier={
                "@id": None
            },
            ethics_assessment={
                "@id": None
            },
            experimental_approach=[{
                "@id": None
            }],
            full_documentation={
                "@id": None
            },
            full_name=self.name if hasattr(self, "name") else None,
            license={
                "@id": None
            },
            release_date=date(1970,1,1),
            short_name=self.name[:30] if hasattr(self, "name") else "",
            technique=[{
                "@id": None
            }],
            version_identifier="",
            version_innovation="",
            description=(self.description or "")[:2000] if hasattr(self, "description") else "",
        )
        return DatasetJsonModel(
            id=metadata.id,
            type=Dataset.get_model_type(),
            metadata=metadata,
            urls=[Url(**url) for url in self.urls]
        )

class DatasetJsonModel(ConfigBaseModel):
    id: str = Field(..., alias="@id")
    type: str = Field(Dataset.get_model_type(), alias="@type", const=True)
    metadata: DatasetVersionModel
    urls: List[Url]

class OriginDescription(Dataset, type_id="fzj/tmp/simpleOriginInfo/v0.0.1"):
    def __init__(self, name, description, urls):
        Dataset.__init__(self, None, description=description)
        # we model the following as property functions,
        # so derived classes may replace them with a lazy loading mechanism.
        self.name = name
        self._urls = urls

    @property
    def urls(self):
        return self._urls

    @classmethod
    def _from_json(cls, spec):
        type_id = cls.extract_type_id(spec)
        assert type_id == cls.type_id
        return cls(
            name=spec["name"],
            description=spec.get("description"),
            urls=spec.get("url", []),
        )


class _EbrainsKgV3Base:
    BASE_URL = "https://core.kg.ebrains.eu/v3-beta/queries"
    QUERY_ID = None
    STAGE = "RELEASED"

    def __init__(self, _id_spec: Dict[str, Any]) -> None:
        self._id_spec = _id_spec
        self._spec = None
        

    @classmethod
    def _query(cls, spec: Dict[str, Any]=None):
        
        # for easy compatibility with data returned by KG
        # KG no longer uses @id for key, but instead id (by default)
        # also id prepends domain information (e.g. https://kg.ebrains.eu/api/instances/{uuid})
        # therefore, also extract the uuid protion

        if spec is not None:
            at_id=spec.get("@id")
            kg_id = spec.get("id")
            kg_id_search = kg_id and re.search(r'[a-f0-9-]+$', kg_id)
        
            assert at_id is not None or kg_id_search is not None
            uuid = at_id or kg_id_search.group()

        assert hasattr(cls, 'type_id')
        
        # for easy compatibility with data returned by KG
        # KG no longer uses @type for key, but type
        # also type is List[str]
        assert spec.get("@type") == cls.type_id or any ([t == cls.type_id for t in spec.get("type", [])])
        
        
        url=f"{cls.BASE_URL}/{cls.QUERY_ID}/instances?stage={cls.STAGE}"
        if spec is not None:
            url += f"&instanceId={uuid}"

        result = EbrainsRequest(url, DECODERS['.json']).get()

        assert 'data' in result

        if spec is not None:
            assert result.get('total') == 1
            assert result.get('size') == 1
            return result.get("data")[0]

        return result.get('data', [])
        

class EbrainsKgV3Dataset(Dataset, _EbrainsKgV3Base, type_id="https://openminds.ebrains.eu/core/Dataset"):
    BASE_URL = "https://core.kg.ebrains.eu/v3-beta/queries"
    QUERY_ID = "138111f9-1aa4-43f5-8e0a-6e6ed085fa3e"

    def __init__(self, spec: Dict[str, Any]):
        super().__init__(None)
        found = re.search(r'[a-f0-9-]+$', spec.get('id'))
        assert found
        self.id = found.group()
        self._description_cached = spec.get("description")
        self._spec = spec

    @classmethod
    def _from_json(cls, spec: Dict[str, Any]):
        json_obj = cls._query(spec)
        return cls(json_obj)


class EbrainsKgV3DatasetVersion(Dataset, _EbrainsKgV3Base, type_id="https://openminds.ebrains.eu/core/DatasetVersion"):
    
    BASE_URL = "https://core.kg.ebrains.eu/v3-beta/queries"
    QUERY_ID = "f7489d01-2f90-410c-9812-9ee7d10cc5be"

    def __init__(self, _id_spec: Dict[str, Any]):
        _EbrainsKgV3Base.__init__(self, _id_spec)
        Dataset.__init__(self, None)

    @classmethod
    def _from_json(cls, spec: Dict[str, Any]):
        return cls(spec)
    
    @property
    def description(self):
        if self._spec is None:
            self._spec = self._query(self._id_spec)
        
        self._description_cached = self._spec.get("description")
        
        if self._description_cached is not None and self._description_cached != '':
            return self._description_cached

        parent_datasets = self._spec.get("belongsTo", [])
        if len(parent_datasets) == 0:
            return None
        if len(parent_datasets) > 1:
            logger.warn(f"EbrainsKgV3DatasetVersion.description: more than one parent dataset found. Using the first one...")

        parent = EbrainsKgV3Dataset._from_json(parent_datasets[0])
        return parent.description



class EbrainsDataset(Dataset, type_id="minds/core/dataset/v1.0.0"):
    def __init__(self, id, name, embargo_status=None):
        Dataset.__init__(self, id, description=None)
        self.embargo_status = embargo_status
        self._name_cached = name
        self._detail = None
        if id is None:
            raise TypeError("Dataset id is required")

        match = re.search(r"([a-f0-9-]+)$", id)
        if not match:
            raise ValueError(
                f"{self.__class__.__name__} initialized with invalid id: {self.id}"
            )
        self._detail_loader = EbrainsKgQuery(
            query_id="interactiveViewerKgQuery-v1_0",
            instance_id=match.group(1),
            params={"vocab": "https://schema.hbp.eu/myQuery/"},
        )

    @property
    def detail(self):
        return self._detail_loader.data

    @property
    def name(self):
        if self._name_cached is None:
            self._name_cached = self.detail.get("name")
        return self._name_cached

    @property
    def publications(self):
        return self.detail.get('publications')

    @property
    def urls(self):
        return [
            {
                "doi": f,
            }
            for f in self.detail.get("kgReference", [])
        ]

    @property
    def description(self):
        return self.detail.get("description")

    @property
    def contributors(self):
        return self.detail.get("contributors")

    @property
    def ebrains_page(self):
        return f"https://search.kg.ebrains.eu/instances/{self.id}"

    @property
    def custodians(self):
        return self.detail.get("custodians")

    def __hash__(self):
        return hash(self.id)

    def __eq__(self, o: object) -> bool:
        if type(o) is not EbrainsDataset and not issubclass(type(o), EbrainsDataset):
            return False
        return self.id == o.id

    @classmethod
    def _from_json(cls, spec):
        type_id = cls.extract_type_id(spec)
        assert type_id == cls.type_id
        return cls(
            id=spec.get("kgId"),
            name=spec.get("name"),
            embargo_status=spec.get("embargo_status", None),
        )



# generated by datamodel-codegen:
#   filename:  https://raw.githubusercontent.com/HumanBrainProject/openMINDS/3fa86f956b407b2debf47c2e1b6314e37579c707/v3/core/v4/miscellaneous/funding.schema.json

from typing import Any, Dict, Optional

from pydantic import Field
from siibra.openminds.base import SiibraBaseModel


class Model(SiibraBaseModel):
    id: str = Field(..., alias='@id', description='Metadata node identifier.')
    type: str = Field(..., alias='@type')
    acknowledgement: Optional[str] = Field(
        None,
        alias='acknowledgement',
        description='Offical declaration or avowal of appreciation of an act or achievement.',
        title='acknowledgement',
    )
    award_number: Optional[str] = Field(
        None,
        alias='awardNumber',
        description='Machine-readable identifier for a benefit that is conferred or bestowed on the basis of merit or need.',
        title='awardNumber',
    )
    award_title: Optional[str] = Field(
        None,
        alias='awardTitle',
        description='Human-readable identifier for a benefit that is conferred or bestowed on the basis of merit or need.',
        title='awardTitle',
    )
    funder: Dict[str, Any] = Field(
        ...,
        alias='funder',
        description='Legal person that provides money for a particular purpose.',
        title='funder',
    )



# generated by datamodel-codegen:
#   filename:  https://raw.githubusercontent.com/HumanBrainProject/openMINDS/3fa86f956b407b2debf47c2e1b6314e37579c707/v3/core/v4/data/file.schema.json

from typing import Any, Dict, List, Optional

from pydantic import Field
from siibra.openminds.base import SiibraBaseModel


class Hash(SiibraBaseModel):
    algorithm: str = Field(
        ...,
        description='Procedure for solving a mathematical problem in a finite number of steps. Can involve repetition of an operation.',
        title='algorithm',
    )
    digest: str = Field(
        ...,
        description='Summation or condensation of a body of information.',
        title='digest',
    )


class StorageSize(SiibraBaseModel):
    type_of_uncertainty: Optional[Any] = Field(
        None,
        alias='typeOfUncertainty',
        description='Distinct technique used to quantify the uncertainty of a measurement.',
        title='typeOfUncertainty',
    )
    uncertainty: Optional[List[float]] = Field(
        None,
        description='Quantitative value range defining the uncertainty of a measurement.',
        max_items=2,
        min_items=2,
        title='uncertainty',
    )
    unit: Optional[Any] = Field(
        None,
        description='Determinate quantity adopted as a standard of measurement.',
        title='unit',
    )
    value: float = Field(..., description='Entry for a property.', title='value')


class Model(SiibraBaseModel):
    id: str = Field(..., alias='@id', description='Metadata node identifier.')
    type: str = Field(..., alias='@type')
    iri: str = Field(
        ...,
        alias='IRI',
        description='Stands for Internationalized Resource Identifier which is an internet protocol standard that builds on the URI protocol, extending the set of permitted characters to include Unicode/ISO 10646.',
        title='IRI',
    )
    content_description: Optional[str] = Field(
        None,
        alias='contentDescription',
        title='contentDescription',
    )
    data_type: Optional[List[Any]] = Field(
        None,
        alias='dataType',
        min_items=1,
        title='dataType',
    )
    file_repository: Optional[Dict[str, Any]] = Field(
        None,
        alias='fileRepository',
        title='fileRepository',
    )
    format: Optional[Dict[str, Any]] = Field(
        None,
        alias='format',
        description='Method of digitally organizing and structuring data or information.',
        title='format',
    )
    hash: Optional['Hash'] = Field(
        None,
        alias='hash',
        description='Structured information on a hash.',
    )
    is_part_of: Optional[List[Any]] = Field(
        None,
        alias='isPartOf',
        description='Reference to the ensemble of multiple things or beings.',
        min_items=1,
        title='isPartOf',
    )
    name: str = Field(
        ...,
        alias='name',
        description='Word or phrase that constitutes the distinctive designation of a being or thing.',
        title='name',
    )
    special_usage_role: Optional[Dict[str, Any]] = Field(
        None,
        alias='specialUsageRole',
        description='Particular function of something when it is used.',
        title='specialUsageRole',
    )
    storage_size: Optional['StorageSize'] = Field(
        None,
        alias='storageSize',
        description='Structured information on a quantitative value.',
    )



# generated by datamodel-codegen:
#   filename:  https://raw.githubusercontent.com/HumanBrainProject/openMINDS/3fa86f956b407b2debf47c2e1b6314e37579c707/v3/core/v4/data/license.schema.json

from typing import List, Optional

from pydantic import Field
from siibra.openminds.base import SiibraBaseModel


class Model(SiibraBaseModel):
    id: str = Field(..., alias='@id', description='Metadata node identifier.')
    type: str = Field(..., alias='@type')
    full_name: str = Field(
        ...,
        alias='fullName',
        description='Whole, non-abbreviated name of something or somebody.',
        title='fullName',
    )
    legal_code: str = Field(
        ...,
        alias='legalCode',
        description='Type of legislation that claims to cover the law system (complete or parts) as it existed at the time the code was enacted.',
        title='legalCode',
    )
    short_name: str = Field(
        ...,
        alias='shortName',
        description='Shortened or fully abbreviated name of something or somebody.',
        title='shortName',
    )
    webpage: Optional[List[str]] = Field(
        None,
        alias='webpage',
        description='Hypertext document (block of information) found on the World Wide Web.',
        min_items=1,
        title='webpage',
    )



# generated by datamodel-codegen:
#   filename:  https://raw.githubusercontent.com/HumanBrainProject/openMINDS/3fa86f956b407b2debf47c2e1b6314e37579c707/v3/core/v4/research/strain.schema.json

from typing import Any, Dict, List, Optional

from pydantic import Field, constr
from siibra.openminds.base import SiibraBaseModel


class StockNumber(SiibraBaseModel):
    identifier: str = Field(
        ...,
        description='Term or code used to identify something or someone.',
        title='identifier',
    )
    vendor: Any = Field(..., title='vendor')


class Model(SiibraBaseModel):
    id: str = Field(..., alias='@id', description='Metadata node identifier.')
    type: str = Field(..., alias='@type')
    rrid: Optional[Dict[str, Any]] = Field(
        None, alias='RRID', title='RRID'
    )
    background_strain: Optional[List[Any]] = Field(
        None,
        alias='backgroundStrain',
        max_items=2,
        min_items=1,
        title='backgroundStrain',
    )
    breeding_type: Optional[Dict[str, Any]] = Field(
        None,
        alias='breedingType',
        title='breedingType',
    )
    description: Optional[str] = Field(
        None,
        alias='description',
        description='Longer statement or account giving the characteristics of someone or something.',
        title='description',
    )
    disease_model: Optional[List[Any]] = Field(
        None,
        alias='diseaseModel',
        min_items=1,
        title='diseaseModel',
    )
    genetic_strain_type: Dict[str, Any] = Field(
        ...,
        alias='geneticStrainType',
        title='geneticStrainType',
    )
    identifier: Optional[List[str]] = Field(
        None,
        alias='identifier',
        description='Term or code used to identify something or someone.',
        min_items=1,
        title='identifier',
    )
    laboratory_code: Optional[constr(regex=r'^[A-Z]([a-z]?)+$')] = Field(
        None,
        alias='laboratoryCode',
        title='laboratoryCode',
    )
    name: str = Field(
        ...,
        alias='name',
        description='Word or phrase that constitutes the distinctive designation of a being or thing.',
        title='name',
    )
    ontology_identifier: Optional[List[str]] = Field(
        None,
        alias='ontologyIdentifier',
        description='Term or code used to identify something or someone registered within a particular ontology.',
        min_items=1,
        title='ontologyIdentifier',
    )
    phenotype: Optional[str] = Field(
        None,
        alias='phenotype',
        description='Physical expression of one or more genes of an organism.',
        title='phenotype',
    )
    species: Optional[Dict[str, Any]] = Field(
        None,
        alias='species',
        description='Category of biological classification comprising related organisms or populations potentially capable of interbreeding, and being designated by a binomial that consists of the name of a genus followed by a Latin or latinized uncapitalized noun or adjective.',
        title='species',
    )
    stock_number: Optional['StockNumber'] = Field(
        None, alias='stockNumber'
    )
    synonym: Optional[List[str]] = Field(
        None,
        alias='synonym',
        description='Words or expressions used in the same language that have the same or nearly the same meaning in some or all senses.',
        min_items=1,
        title='synonym',
    )



# generated by datamodel-codegen:
#   filename:  https://raw.githubusercontent.com/HumanBrainProject/openMINDS/3fa86f956b407b2debf47c2e1b6314e37579c707/v3/SANDS/v3/atlas/parcellationEntityVersion.schema.json

from typing import Any, List, Optional, Union

from pydantic import Field, constr
from siibra.openminds.base import SiibraBaseModel


class Coordinates(SiibraBaseModel):
    type_of_uncertainty: Optional[Any] = Field(
        None,
        alias='typeOfUncertainty',
        description='Distinct technique used to quantify the uncertainty of a measurement.',
        title='typeOfUncertainty',
    )
    uncertainty: Optional[List[float]] = Field(
        None,
        description='Quantitative value range defining the uncertainty of a measurement.',
        title='uncertainty',
    )
    unit: Optional[Any] = Field(
        None,
        description='Determinate quantity adopted as a standard of measurement.',
        title='unit',
    )
    value: float = Field(..., description='Entry for a property.', title='value')


class BestViewPoint(SiibraBaseModel):
    coordinate_space: Any = Field(
        ...,
        alias='coordinateSpace',
        description='Two or three dimensional geometric setting.',
        title='coordinateSpace',
    )
    coordinates: List['Coordinates'] = Field(
        ..., description='Structured information on a quantitative value.'
    )


class HasAnnotation(SiibraBaseModel):
    best_view_point: Optional['BestViewPoint'] = Field(
        None,
        alias='bestViewPoint',
        description='Structured information on a coordinate point.',
    )
    criteria: Optional[Any] = Field(
        None,
        description='Aspects or standards on which a judgement or decision is based.',
        title='criteria',
    )
    criteria_quality_type: Any = Field(
        ...,
        alias='criteriaQualityType',
        description='Distinct class that defines how the judgement or decision was made for a particular criteria.',
        title='criteriaQualityType',
    )
    display_color: Optional[constr(regex=r'^#[0-9A-Fa-f]{6}$')] = Field(
        None,
        alias='displayColor',
        description='Preferred coloring.',
        title='displayColor',
    )
    inspired_by: Optional[List] = Field(
        None,
        alias='inspiredBy',
        description='Reference to an inspiring element.',
        title='inspiredBy',
    )
    internal_identifier: str = Field(
        ...,
        alias='internalIdentifier',
        description='Term or code that identifies someone or something within a particular product.',
        title='internalIdentifier',
    )
    laterality: Optional[List] = Field(
        None,
        description='Differentiation between a pair of lateral homologous parts of the body.',
        title='laterality',
    )
    visualized_in: Optional[Any] = Field(
        None,
        alias='visualizedIn',
        description='Reference to an image in which something is visible.',
        title='visualizedIn',
    )


class RelationAssessmentItem(SiibraBaseModel):
    criteria: Optional[Any] = Field(
        None,
        description='Aspects or standards on which a judgement or decision is based.',
        title='criteria',
    )
    in_relation_to: Any = Field(
        ...,
        alias='inRelationTo',
        description='Reference to a related element.',
        title='inRelationTo',
    )
    qualitative_overlap: Any = Field(
        ...,
        alias='qualitativeOverlap',
        description='Semantic characterization of how much two things occupy the same space.',
        title='qualitativeOverlap',
    )


class QuantitativeOverlapItem(SiibraBaseModel):
    type_of_uncertainty: Optional[Any] = Field(
        None,
        alias='typeOfUncertainty',
        description='Distinct technique used to quantify the uncertainty of a measurement.',
        title='typeOfUncertainty',
    )
    uncertainty: Optional[List[float]] = Field(
        None,
        description='Quantitative value range defining the uncertainty of a measurement.',
        title='uncertainty',
    )
    unit: Optional[Any] = Field(
        None,
        description='Determinate quantity adopted as a standard of measurement.',
        title='unit',
    )
    value: float = Field(..., description='Entry for a property.', title='value')


class QuantitativeOverlapItem1(SiibraBaseModel):
    max_value: float = Field(
        ...,
        alias='maxValue',
        description='Greatest quantity attained or allowed.',
        title='maxValue',
    )
    max_value_unit: Optional[Any] = Field(
        None, alias='maxValueUnit', title='maxValueUnit'
    )
    min_value: float = Field(
        ...,
        alias='minValue',
        description='Smallest quantity attained or allowed.',
        title='minValue',
    )
    min_value_unit: Optional[Any] = Field(
        None, alias='minValueUnit', title='minValueUnit'
    )


class RelationAssessmentItem1(SiibraBaseModel):
    criteria: Optional[Any] = Field(
        None,
        description='Aspects or standards on which a judgement or decision is based.',
        title='criteria',
    )
    in_relation_to: Any = Field(
        ...,
        alias='inRelationTo',
        description='Reference to a related element.',
        title='inRelationTo',
    )
    quantitative_overlap: Union[
        'QuantitativeOverlapItem', 'QuantitativeOverlapItem1'
    ] = Field(..., alias='quantitativeOverlap')


class Model(SiibraBaseModel):
    id: str = Field(..., alias='@id', description='Metadata node identifier.')
    type: str = Field(..., alias='@type')
    has_annotation: Optional['HasAnnotation'] = Field(
        None, alias='hasAnnotation'
    )
    has_parent: Optional[List[Any]] = Field(
        None,
        alias='hasParent',
        description='Reference to a parent object or legal person.',
        title='hasParent',
    )
    lookup_label: Optional[str] = Field(
        None,
        alias='lookupLabel',
        title='lookupLabel',
    )
    name: Optional[str] = Field(
        None,
        alias='name',
        description='Word or phrase that constitutes the distinctive designation of a being or thing.',
        title='name',
    )
    ontology_identifier: Optional[List[str]] = Field(
        None,
        alias='ontologyIdentifier',
        description='Term or code used to identify something or someone registered within a particular ontology.',
        title='ontologyIdentifier',
    )
    relation_assessment: Optional[
        Union['RelationAssessmentItem', 'RelationAssessmentItem1']
    ] = Field(None, alias='relationAssessment')
    version_identifier: str = Field(
        ...,
        alias='versionIdentifier',
        description='Term or code used to identify the version of something.',
        title='versionIdentifier',
    )
    version_innovation: Optional[str] = Field(
        None,
        alias='versionInnovation',
        description='Documentation on what changed in comparison to a previously published form of something.',
        title='versionInnovation',
    )



# generated by datamodel-codegen:
#   filename:  https://raw.githubusercontent.com/HumanBrainProject/openMINDS/3fa86f956b407b2debf47c2e1b6314e37579c707/v3/core/v4/miscellaneous/ISBN.schema.json

from pydantic import Field, constr
from siibra.openminds.base import SiibraBaseModel


class Model(SiibraBaseModel):
    id: str = Field(..., alias='@id', description='Metadata node identifier.')
    type: str = Field(..., alias='@type')
    identifier: constr(
        regex=r'^([0-9]{3}-|)[0-9]{1}-[0-9]{3}-[0-9]{5}-[0-9]{1}$'
    ) = Field(
        ...,
        alias='identifier',
        description='Term or code used to identify something or someone.',
        title='identifier',
    )



# generated by datamodel-codegen:
#   filename:  https://raw.githubusercontent.com/HumanBrainProject/openMINDS/3fa86f956b407b2debf47c2e1b6314e37579c707/v3/core/v4/data/hash.schema.json

from pydantic import Field
from siibra.openminds.base import SiibraBaseModel


class Model(SiibraBaseModel):
    id: str = Field(..., alias='@id', description='Metadata node identifier.')
    type: str = Field(..., alias='@type')
    algorithm: str = Field(
        ...,
        alias='algorithm',
        description='Procedure for solving a mathematical problem in a finite number of steps. Can involve repetition of an operation.',
        title='algorithm',
    )
    digest: str = Field(
        ...,
        alias='digest',
        description='Summation or condensation of a body of information.',
        title='digest',
    )



# generated by datamodel-codegen:
#   filename:  https://raw.githubusercontent.com/HumanBrainProject/openMINDS/3fa86f956b407b2debf47c2e1b6314e37579c707/v3/core/v4/products/project.schema.json

from typing import Any, Dict, List, Optional

from pydantic import Field
from siibra.openminds.base import SiibraBaseModel


class Model(SiibraBaseModel):
    id: str = Field(..., alias='@id', description='Metadata node identifier.')
    type: str = Field(..., alias='@type')
    coordinator: Optional[List[Any]] = Field(
        None,
        alias='coordinator',
        description='Legal person who organizes the collaborative work of people or groups.',
        min_items=1,
        title='coordinator',
    )
    description: str = Field(
        ...,
        alias='description',
        description='Longer statement or account giving the characteristics of someone or something.',
        title='description',
    )
    full_name: str = Field(
        ...,
        alias='fullName',
        description='Whole, non-abbreviated name of something or somebody.',
        title='fullName',
    )
    has_research_products: List[Any] = Field(
        ...,
        alias='hasResearchProducts',
        description='Reference to subsidiary research products.',
        min_items=2,
        title='hasResearchProducts',
    )
    homepage: Optional[Dict[str, Any]] = Field(
        None,
        alias='homepage',
        description='Main website of something or someone.',
        title='homepage',
    )
    short_name: str = Field(
        ...,
        alias='shortName',
        description='Shortened or fully abbreviated name of something or somebody.',
        title='shortName',
    )



# generated by datamodel-codegen:
#   filename:  https://raw.githubusercontent.com/HumanBrainProject/openMINDS/3fa86f956b407b2debf47c2e1b6314e37579c707/v3/computation/v1/optimization.schema.json

from typing import Any, Dict, List, Optional, Union

from pydantic import Field
from siibra.openminds.base import SiibraBaseModel


class ResourceUsageItem(SiibraBaseModel):
    type_of_uncertainty: Optional[Any] = Field(
        None,
        alias='typeOfUncertainty',
        description='Distinct technique used to quantify the uncertainty of a measurement.',
        title='typeOfUncertainty',
    )
    uncertainty: Optional[List[float]] = Field(
        None,
        description='Quantitative value range defining the uncertainty of a measurement.',
        max_items=2,
        min_items=2,
        title='uncertainty',
    )
    unit: Optional[Any] = Field(
        None,
        description='Determinate quantity adopted as a standard of measurement.',
        title='unit',
    )
    value: float = Field(..., description='Entry for a property.', title='value')


class ResourceUsageItem1(SiibraBaseModel):
    max_value: float = Field(
        ...,
        alias='maxValue',
        description='Greatest quantity attained or allowed.',
        title='maxValue',
    )
    max_value_unit: Optional[Any] = Field(
        None, alias='maxValueUnit', title='maxValueUnit'
    )
    min_value: float = Field(
        ...,
        alias='minValue',
        description='Smallest quantity attained or allowed.',
        title='minValue',
    )
    min_value_unit: Optional[Any] = Field(
        None, alias='minValueUnit', title='minValueUnit'
    )


class Model(SiibraBaseModel):
    id: str = Field(..., alias='@id', description='Metadata node identifier.')
    type: str = Field(..., alias='@type')
    environment: Dict[str, Any] = Field(
        ..., alias='environment', title='environment'
    )
    input: Optional[Dict[str, Any]] = Field(
        None,
        alias='input',
        description='Something or someone that is put into or participates in a process or machine.',
        title='input',
    )
    launch_configuration: Dict[str, Any] = Field(
        ...,
        alias='launchConfiguration',
        title='launchConfiguration',
    )
    output: Optional[Dict[str, Any]] = Field(
        None,
        alias='output',
        description='Something or someone that comes out of, is delivered or produced by a process or machine.',
        title='output',
    )
    resource_usage: Optional[Union['ResourceUsageItem', 'ResourceUsageItem1']] = Field(
        None, alias='resourceUsage'
    )
    started_by: Optional[Dict[str, Any]] = Field(
        None, alias='startedBy', title='startedBy'
    )
    status: Optional[Dict[str, Any]] = Field(
        None, alias='status', title='status'
    )
    tags: Optional[List[str]] = Field(
        None, alias='tags', min_items=1, title='tags'
    )
    was_informed_by: Optional[Dict[str, Any]] = Field(
        None,
        alias='wasInformedBy',
        title='wasInformedBy',
    )



# generated by datamodel-codegen:
#   filename:  https://raw.githubusercontent.com/HumanBrainProject/openMINDS/3fa86f956b407b2debf47c2e1b6314e37579c707/v3/SANDS/v3/atlas/parcellationTerminology.schema.json

from typing import Any, List, Optional

from pydantic import Field
from siibra.openminds.base import SiibraBaseModel


class Model(SiibraBaseModel):
    id: str = Field(..., alias='@id', description='Metadata node identifier.')
    type: str = Field(..., alias='@type')
    defined_in: Optional[List[Any]] = Field(
        None,
        alias='definedIn',
        description='Reference to a file instance in which something is stored.',
        min_items=1,
        title='definedIn',
    )
    has_entity: List[Any] = Field(
        ...,
        alias='hasEntity',
        min_items=1,
        title='hasEntity',
    )
    ontology_identifier: Optional[List[str]] = Field(
        None,
        alias='ontologyIdentifier',
        description='Term or code used to identify something or someone registered within a particular ontology.',
        min_items=1,
        title='ontologyIdentifier',
    )



# generated by datamodel-codegen:
#   filename:  https://raw.githubusercontent.com/HumanBrainProject/openMINDS/3fa86f956b407b2debf47c2e1b6314e37579c707/v3/core/v4/data/filePathPattern.schema.json

from typing import Any, Dict

from pydantic import Field
from siibra.openminds.base import SiibraBaseModel


class Model(SiibraBaseModel):
    id: str = Field(..., alias='@id', description='Metadata node identifier.')
    type: str = Field(..., alias='@type')
    grouping_type: Dict[str, Any] = Field(
        ...,
        alias='groupingType',
        title='groupingType',
    )
    regex: str = Field(
        ..., alias='regex', title='regex'
    )



# generated by datamodel-codegen:
#   filename:  https://raw.githubusercontent.com/HumanBrainProject/openMINDS/3fa86f956b407b2debf47c2e1b6314e37579c707/v3/core/v4/miscellaneous/URL.schema.json

from pydantic import Field
from siibra.openminds.base import SiibraBaseModel


class Model(SiibraBaseModel):
    id: str = Field(..., alias='@id', description='Metadata node identifier.')
    type: str = Field(..., alias='@type')
    url: str = Field(..., alias='URL', title='URL')



# generated by datamodel-codegen:
#   filename:  https://raw.githubusercontent.com/HumanBrainProject/openMINDS/3fa86f956b407b2debf47c2e1b6314e37579c707/v3/core/v4/data/fileRepositoryStructure.schema.json

from typing import Any, Optional

from pydantic import Field
from siibra.openminds.base import SiibraBaseModel


class FilePathPattern(SiibraBaseModel):
    grouping_type: Any = Field(..., alias='groupingType', title='groupingType')
    regex: str = Field(..., title='regex')


class Model(SiibraBaseModel):
    id: str = Field(..., alias='@id', description='Metadata node identifier.')
    type: str = Field(..., alias='@type')
    file_path_pattern: 'FilePathPattern' = Field(
        ..., alias='filePathPattern'
    )
    lookup_label: Optional[str] = Field(
        None,
        alias='lookupLabel',
        title='lookupLabel',
    )



# generated by datamodel-codegen:
#   filename:  https://raw.githubusercontent.com/HumanBrainProject/openMINDS/3fa86f956b407b2debf47c2e1b6314e37579c707/v3/core/v4/research/protocolExecution.schema.json

from typing import Any, Dict, List, Optional, Union

from pydantic import Field
from siibra.openminds.base import SiibraBaseModel


class ValueItem(SiibraBaseModel):
    type_of_uncertainty: Optional[Any] = Field(
        None,
        alias='typeOfUncertainty',
        description='Distinct technique used to quantify the uncertainty of a measurement.',
        title='typeOfUncertainty',
    )
    uncertainty: Optional[List[float]] = Field(
        None,
        description='Quantitative value range defining the uncertainty of a measurement.',
        max_items=2,
        min_items=2,
        title='uncertainty',
    )
    unit: Optional[Any] = Field(
        None,
        description='Determinate quantity adopted as a standard of measurement.',
        title='unit',
    )
    value: float = Field(..., description='Entry for a property.', title='value')


class ValueItem1(SiibraBaseModel):
    max_value: float = Field(
        ...,
        alias='maxValue',
        description='Greatest quantity attained or allowed.',
        title='maxValue',
    )
    max_value_unit: Optional[Any] = Field(
        None, alias='maxValueUnit', title='maxValueUnit'
    )
    min_value: float = Field(
        ...,
        alias='minValue',
        description='Smallest quantity attained or allowed.',
        title='minValue',
    )
    min_value_unit: Optional[Any] = Field(
        None, alias='minValueUnit', title='minValueUnit'
    )


class ParameterItem(SiibraBaseModel):
    name: str = Field(
        ...,
        description='Word or phrase that constitutes the distinctive designation of a being or thing.',
        title='name',
    )
    value: Union['ValueItem', 'ValueItem1']


class ParameterItem1(SiibraBaseModel):
    name: str = Field(
        ...,
        description='Word or phrase that constitutes the distinctive designation of a being or thing.',
        title='name',
    )
    value: str = Field(..., description='Entry for a property.', title='value')


class ParameterSet(SiibraBaseModel):
    context: str = Field(..., title='context')
    parameter: Union['ParameterItem', 'ParameterItem1']
    relevant_for: Any = Field(
        ...,
        alias='relevantFor',
        description='Reference to what or whom something or someone bears siginificance.',
        title='relevantFor',
    )


class Model(SiibraBaseModel):
    id: str = Field(..., alias='@id', description='Metadata node identifier.')
    type: str = Field(..., alias='@type')
    behavioral_protocol: Optional[List[Any]] = Field(
        None,
        alias='behavioralProtocol',
        min_items=1,
        title='behavioralProtocol',
    )
    description: Optional[str] = Field(
        None,
        alias='description',
        description='Longer statement or account giving the characteristics of someone or something.',
        title='description',
    )
    ended_at_time: Optional[Union[Any, Any]] = Field(
        None,
        alias='endedAtTime',
        title='endedAtTime',
    )
    input: List[Any] = Field(
        ...,
        alias='input',
        description='Something or someone that is put into or participates in a process or machine.',
        min_items=1,
        title='input',
    )
    is_part_of: Dict[str, Any] = Field(
        ...,
        alias='isPartOf',
        description='Reference to the ensemble of multiple things or beings.',
        title='isPartOf',
    )
    lookup_label: Optional[str] = Field(
        None,
        alias='lookupLabel',
        title='lookupLabel',
    )
    output: List[Any] = Field(
        ...,
        alias='output',
        description='Something or someone that comes out of, is delivered or produced by a process or machine.',
        min_items=1,
        title='output',
    )
    parameter_set: Optional['ParameterSet'] = Field(
        None,
        alias='parameterSet',
        description='Structured information on a used parameter set.',
    )
    preparation_design: Optional[Dict[str, Any]] = Field(
        None,
        alias='preparationDesign',
        title='preparationDesign',
    )
    protocol: List[Any] = Field(
        ...,
        alias='protocol',
        description='Plan that describes the process of a scientific or medical experiment, treatment, or procedure.',
        min_items=1,
        title='protocol',
    )
    started_at_time: Optional[Union[Any, Any]] = Field(
        None,
        alias='startedAtTime',
        title='startedAtTime',
    )
    study_target: Optional[List[Any]] = Field(
        None,
        alias='studyTarget',
        description='Structure or function that was targeted within a study.',
        min_items=1,
        title='studyTarget',
    )



from abc import ABC, abstractclassmethod, abstractmethod, abstractproperty
import numpy as np
import zlib
import base64
from pydantic import Field

from ..openminds.base import ConfigBaseModel

class JSONSerializable(ABC):

    @abstractclassmethod
    def get_model_type(Cls) -> str:
        """
        The get_modal_type abstract class method should populate the @type attribute of the model returned by to_model
        It should also allow unified indexing of a list of JSONSerializable according to types without first calling to_model method (which could be expensive)

        e.g.

        without get_modal_type abstract class method:
        # filter features based on type
        found_feature = [feat for feat in features if feat.to_model().type == feature_id]

        with get_modal_type abstract class method:
        # finding a feature based on feature_id
        found_feature = [feat for feat in features if feat.get_modal_type() == feature_id]
        """
        raise AttributeError("get_model_type class method needs to be overwritten by subclass!")

    @abstractproperty
    def model_id(self) -> str:
        """
        The model_id abstract property should populate the @id attribute of the model returned by to_model
        It should also allow unified indexing of a list of JSONSerializable without first calling to_model method (which could be expensive)

        e.g.

        without model_id abstract property:
        # finding a feature based on feature_id
        found_feature = [feat for feat in features if feat.model_id == feature_id]

        with model_id abstract property:
        # finding a feature based on feature_id
        found_feature = [feat for feat in features if feat.model_id == feature_id]
        """
        raise AttributeError("model_id property needs to be overwritten by subclass!")

    @abstractmethod
    def to_model(self, **kwargs):
        raise AttributeError("JSONSerializable needs to have to_model overwritten")
    
    def __init_subclass__(cls) -> None:
        assert cls.to_model.__annotations__.get("return"), f"{cls}: JSONSerializable to_model overwritten method needs to be typed."
        return super().__init_subclass__()


class NpArrayDataModel(ConfigBaseModel):
    
    content_type: str = Field("application/octet-stream")
    content_encoding: str = Field("gzip; base64")
    x_width: int = Field(..., alias="x-width")
    x_height: int = Field(..., alias="x-height")
    x_channel: int = Field(..., alias="x-channel")
    dtype: str
    content: str

    def __init__(self, np_data=None, **data) -> None:

        if np_data is None:
            return super().__init__(**data)

        # try to avoid 64 bit any number
        supported_dtype = [
            np.dtype("uint8"),
            np.dtype("int32"),
            np.dtype("float32"),
        ]
        assert type(np_data) is np.ndarray, f"expect input to be numpy array"
        assert np_data.dtype in supported_dtype, f"can only serialize {','.join([str(dtype) for dtype in supported_dtype])} for now"
        assert len(np_data.shape) <= 3, f"can only deal np array with up to 3 dimension"
        
        x_channel = 1 if len(np_data.shape) < 3 else np_data.shape[2]
        x_width = np_data.shape[0]
        x_height = 1 if len(np_data.shape) < 2 else np_data.shape[1]
        dtype = str(np_data.dtype)
        content = base64.b64encode(zlib.compress(np_data.tobytes(order="F")))

        super().__init__(
            x_width=x_width,
            x_height=x_height,
            dtype=dtype,
            content=content,
            x_channel=x_channel,
        )


# generated by datamodel-codegen:
#   filename:  https://raw.githubusercontent.com/HumanBrainProject/openMINDS/3fa86f956b407b2debf47c2e1b6314e37579c707/v3/controlledTerms/v1/operatingSystem.schema.json

from typing import List, Optional

from pydantic import Field
from siibra.openminds.base import SiibraBaseModel


class Model(SiibraBaseModel):
    id: str = Field(..., alias='@id', description='Metadata node identifier.')
    type: str = Field(..., alias='@type')
    definition: Optional[str] = Field(
        None,
        alias='definition',
        description='Short, but precise statement of the meaning of a word, word group, sign or a symbol.',
        title='definition',
    )
    description: Optional[str] = Field(
        None,
        alias='description',
        description='Longer statement or account giving the characteristics of someone or something.',
        title='description',
    )
    interlex_identifier: Optional[str] = Field(
        None,
        alias='interlexIdentifier',
        description='Persistent identifier for a term registered in the InterLex project.',
        title='interlexIdentifier',
    )
    knowledge_space_link: Optional[str] = Field(
        None,
        alias='knowledgeSpaceLink',
        description='Persistent link to an encyclopedia entry in the Knowledge Space project.',
        title='knowledgeSpaceLink',
    )
    name: str = Field(
        ...,
        alias='name',
        description='Word or phrase that constitutes the distinctive designation of a being or thing.',
        title='name',
    )
    preferred_ontology_identifier: Optional[str] = Field(
        None,
        alias='preferredOntologyIdentifier',
        description='Persistent identifier of a preferred ontological term.',
        title='preferredOntologyIdentifier',
    )
    synonym: Optional[List[str]] = Field(
        None,
        alias='synonym',
        description='Words or expressions used in the same language that have the same or nearly the same meaning in some or all senses.',
        min_items=1,
        title='synonym',
    )



# generated by datamodel-codegen:
#   filename:  https://raw.githubusercontent.com/HumanBrainProject/openMINDS/3fa86f956b407b2debf47c2e1b6314e37579c707/v3/core/v4/research/behavioralProtocol.schema.json

from typing import Any, List, Optional

from pydantic import Field
from siibra.openminds.base import SiibraBaseModel


class Model(SiibraBaseModel):
    id: str = Field(..., alias='@id', description='Metadata node identifier.')
    type: str = Field(..., alias='@type')
    description: str = Field(
        ...,
        alias='description',
        description='Longer statement or account giving the characteristics of someone or something.',
        title='description',
    )
    internal_identifier: Optional[str] = Field(
        None,
        alias='internalIdentifier',
        description='Term or code that identifies someone or something within a particular product.',
        title='internalIdentifier',
    )
    name: str = Field(
        ...,
        alias='name',
        description='Word or phrase that constitutes the distinctive designation of a being or thing.',
        title='name',
    )
    stimulation: Optional[List[Any]] = Field(
        None,
        alias='stimulation',
        min_items=1,
        title='stimulation',
    )



from pydantic import BaseModel, Field


class ConfigBaseModel(BaseModel):

    def dict(self, *arg, **kwargs):
        kwargs["by_alias"] = True
        return super().dict(*arg, **kwargs)

    class Config:
        allow_population_by_field_name = True


class VocabModel(ConfigBaseModel):
    vocab: str = Field(..., alias="@vocab")


class SiibraAtIdModel(ConfigBaseModel):
    id: str = Field(..., alias="@id")


class SiibraBaseModel(ConfigBaseModel):
    context: VocabModel = Field(VocabModel(vocab="https://openminds.ebrains.eu/vocab/"), alias="@context")
    



# generated by datamodel-codegen:
#   filename:  https://raw.githubusercontent.com/HumanBrainProject/openMINDS/3fa86f956b407b2debf47c2e1b6314e37579c707/v3/SANDS/v3/atlas/parcellationEntity.schema.json

from typing import Any, Dict, List, Optional

from pydantic import Field
from siibra.openminds.base import SiibraBaseModel


class Model(SiibraBaseModel):
    id: str = Field(..., alias='@id', description='Metadata node identifier.')
    type: str = Field(..., alias='@type')
    has_parent: Optional[List[Any]] = Field(
        None,
        alias='hasParent',
        description='Reference to a parent object or legal person.',
        min_items=1,
        title='hasParent',
    )
    has_version: Optional[List[Any]] = Field(
        None,
        alias='hasVersion',
        description='Reference to variants of an original.',
        min_items=1,
        title='hasVersion',
    )
    lookup_label: Optional[str] = Field(
        None,
        alias='lookupLabel',
        title='lookupLabel',
    )
    name: str = Field(
        ...,
        alias='name',
        description='Word or phrase that constitutes the distinctive designation of a being or thing.',
        title='name',
    )
    ontology_identifier: Optional[List[str]] = Field(
        None,
        alias='ontologyIdentifier',
        description='Term or code used to identify something or someone registered within a particular ontology.',
        min_items=1,
        title='ontologyIdentifier',
    )
    related_uberon_term: Optional[Dict[str, Any]] = Field(
        None,
        alias='relatedUBERONTerm',
        title='relatedUBERONTerm',
    )



# generated by datamodel-codegen:
#   filename:  https://raw.githubusercontent.com/HumanBrainProject/openMINDS/3fa86f956b407b2debf47c2e1b6314e37579c707/v3/core/v4/miscellaneous/quantitativeValueRange.schema.json

from typing import Any, Dict, Optional

from pydantic import Field
from siibra.openminds.base import SiibraBaseModel


class Model(SiibraBaseModel):
    id: str = Field(..., alias='@id', description='Metadata node identifier.')
    type: str = Field(..., alias='@type')
    max_value: float = Field(
        ...,
        alias='maxValue',
        description='Greatest quantity attained or allowed.',
        title='maxValue',
    )
    max_value_unit: Optional[Dict[str, Any]] = Field(
        None,
        alias='maxValueUnit',
        title='maxValueUnit',
    )
    min_value: float = Field(
        ...,
        alias='minValue',
        description='Smallest quantity attained or allowed.',
        title='minValue',
    )
    min_value_unit: Optional[Dict[str, Any]] = Field(
        None,
        alias='minValueUnit',
        title='minValueUnit',
    )



# generated by datamodel-codegen:
#   filename:  https://raw.githubusercontent.com/HumanBrainProject/openMINDS/3fa86f956b407b2debf47c2e1b6314e37579c707/v3/computation/v1/softwareAgent.schema.json

from typing import Any, Dict, Optional

from pydantic import Field
from siibra.openminds.base import SiibraBaseModel


class Model(SiibraBaseModel):
    id: str = Field(..., alias='@id', description='Metadata node identifier.')
    type: str = Field(..., alias='@type')
    environment: Optional[Dict[str, Any]] = Field(
        None,
        alias='environment',
        title='environment',
    )
    name: str = Field(
        ...,
        alias='name',
        description='Word or phrase that constitutes the distinctive designation of a being or thing.',
        title='name',
    )
    software: Dict[str, Any] = Field(
        ..., alias='software', title='software'
    )



# generated by datamodel-codegen:
#   filename:  https://raw.githubusercontent.com/HumanBrainProject/openMINDS/3fa86f956b407b2debf47c2e1b6314e37579c707/v3/core/v4/actors/affiliation.schema.json

from datetime import date
from typing import Any, Dict, Optional

from pydantic import Field
from siibra.openminds.base import SiibraBaseModel


class Model(SiibraBaseModel):
    id: str = Field(..., alias='@id', description='Metadata node identifier.')
    type: str = Field(..., alias='@type')
    end_date: Optional[date] = Field(
        None,
        alias='endDate',
        description='Date in the Gregorian calendar at which something terminates in time.',
        title='endDate',
    )
    organization: Dict[str, Any] = Field(
        ...,
        alias='organization',
        description='Legally accountable, administrative and functional structure.',
        title='organization',
    )
    start_date: Optional[date] = Field(
        None,
        alias='startDate',
        description='Date in the Gregorian calendar at which something begins in time',
        title='startDate',
    )



# generated by datamodel-codegen:
#   filename:  https://raw.githubusercontent.com/HumanBrainProject/openMINDS/3fa86f956b407b2debf47c2e1b6314e37579c707/v3/core/v4/data/copyright.schema.json

from typing import Any, List

from pydantic import Field, constr
from siibra.openminds.base import SiibraBaseModel


class Model(SiibraBaseModel):
    id: str = Field(..., alias='@id', description='Metadata node identifier.')
    type: str = Field(..., alias='@type')
    holder: List[Any] = Field(
        ...,
        alias='holder',
        description='Legal person in possession of something.',
        min_items=1,
        title='holder',
    )
    year: constr(regex=r'([0-9]{4})') = Field(
        ...,
        alias='year',
        description='Cycle in the Gregorian calendar specified by a number and comprised of 365 or 366 days divided into 12 months beginning with January and ending with December.',
        title='year',
    )



# generated by datamodel-codegen:
#   filename:  https://raw.githubusercontent.com/HumanBrainProject/openMINDS/3fa86f956b407b2debf47c2e1b6314e37579c707/v3/core/v4/actors/contactInformation.schema.json

from pydantic import EmailStr, Field
from siibra.openminds.base import SiibraBaseModel


class Model(SiibraBaseModel):
    id: str = Field(..., alias='@id', description='Metadata node identifier.')
    type: str = Field(..., alias='@type')
    email: EmailStr = Field(
        ...,
        alias='email',
        description='Address to which or from which an electronic mail can be sent.',
        title='email',
    )



# Copyright 2018-2021
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .feature import ParcellationFeature
from .query import FeatureQuery

from ..commons import logger, QUIET
from ..core.parcellation import Parcellation
from ..retrieval.repositories import GitlabConnector, EbrainsPublicDatasetConnector
from ..core.serializable_concept import NpArrayDataModel, JSONSerializable
from ..openminds.base import ConfigBaseModel

import pandas as pd
import numpy as np
from typing import Dict, List, Optional, Union
import hashlib
from pydantic import Field


class ConnectivityMatrixDataModel(ConfigBaseModel):
    id: str = Field(..., alias="@id")
    type: str = Field(..., alias="@type")
    name: Optional[str]
    description: Optional[str]
    citation: Optional[str]
    authors: Optional[List[str]]
    cohort: Optional[str]
    subject: Optional[str]

    filename: Optional[str]
    dataset_id: Optional[str]

    parcellations: List[Dict[str, str]]
    matrix: Optional[NpArrayDataModel]
    columns: Optional[List[str]]


class ConnectivityMatrix(ParcellationFeature, JSONSerializable):

    """Structural connectivity matrix grouped by a parcellation."""

    def __init__(self, parcellation_id: str, matrixloader, srcinfo):
        """Construct a parcellation-averaged connectivty matrix.

        Arguments
        ---------
        parcellation_id : str
            Id of corresponding parcellation
        matrixloader : func
            Function which loads the matrix as a pandas dataframe
        """
        ParcellationFeature.__init__(self, parcellation_id)
        self._matrix_loader = matrixloader
        self.src_info = srcinfo

    def __getattr__(self, name):
        if name in self.src_info:
            return self.src_info[name]
        else:
            raise AttributeError(f"{self.__class__} has no attribute {name}")

    def __dir__(self):
        return sorted(
            set(
                dir(super(ConnectivityMatrix, self))
                + list(self.__dict__.keys())
                + list(self.src_info.keys())
            )
        )

    @property
    def matrix(self):
        # load and return the matrix
        return self._matrix_loader()

    def get_profile(self, regionspec):
        for p in self.parcellations:
            region = p.decode_region(regionspec)
            return self.matrix[region]

    def __str__(self):
        return ParcellationFeature.__str__(self) + " " + str(self.src_info)

    @classmethod
    def get_model_type(Cls):
        return "siibra/features/connectivity"

    @property
    def model_id(self):
        return f"{self.get_model_type()}/{hashlib.md5(str(self).encode('utf-8')).hexdigest()}"

    def to_model(self, detail=False, **kwargs) -> ConnectivityMatrixDataModel:

        # these properties do not exist on ConnectivityMatrix, but may be populated from ConnectivityMatrix.src_info via __getattr__
        model_dict_from_getattr = {
            key: getattr(self, key) if hasattr(self, key) else None
            for key in [
                "name",
                "description",
                "citation",
                "authors",
                "cohort",
                "subject",
                "filename",
                "dataset_id",
            ]
        }
        base_model = ConnectivityMatrixDataModel(
            id=self.model_id,
            type=self.get_model_type(),
            parcellations=[{
                "@id": parc.model_id,
            } for parc in self.parcellations],
            **model_dict_from_getattr,
        )
        if detail is False:
            return base_model

        from ..core import Region
        dtype_set = {dtype for dtype in self.matrix.dtypes}
        
        if len(dtype_set) == 0:
            raise TypeError(f"dtype is an empty set!")

        force_float = False
        if len(dtype_set) == 1:
            dtype, = list(dtype_set)
            is_int = np.issubdtype(dtype, int)
            is_float = np.issubdtype(dtype, float)
            assert is_int or is_float, f"expect datatype to be subdtype of either int or float, but is neither: {str(dtype)}"

        if len(dtype_set) > 1:
            logger.warning(f"expect only 1 type of data, but got {len(dtype_set)}, will cast everything to float")
            force_float = True
        
        def get_column_name(col: Union[str, Region]) -> str:
            if isinstance(col, str):
                return col
            if isinstance(col, Region):
                return col.name
            raise TypeError(f"matrix column value {col} of instance {col.__class__} can be be converted to str.")

        base_model.columns = [get_column_name(name) for name in self.matrix.columns.values]
        base_model.matrix=NpArrayDataModel(self.matrix.to_numpy(dtype="float32" if force_float or is_float else "int32"))
        return base_model


class StreamlineCounts(ConnectivityMatrix):
    """Structural connectivity matrix of streamline counts grouped by a parcellation."""

    @classmethod
    def get_model_type(Cls):
        return "siibra/features/connectivity/streamlineCounts"

    def __init__(self, parcellation_id: str, matrixloader, srcinfo):
        super().__init__(parcellation_id, matrixloader, srcinfo)


class StreamlineLengths(ConnectivityMatrix):
    """Structural connectivity matrix of streamline lengths grouped by a parcellation."""
    
    @classmethod
    def get_model_type(Cls):
        return "siibra/features/connectivity/streamlineLengths"
    
    def __init__(self, parcellation_id: str, matrixloader, srcinfo):
        super().__init__(parcellation_id, matrixloader, srcinfo)


class FunctionalConnectivity(ConnectivityMatrix):
    """Functional connectivity matrix, grouped by a parcellation."""

    @classmethod
    def get_model_type(Cls):
        return "siibra/features/connectivity/functional"

    def __init__(self, parcellation_id: str, matrixloader, paradigm: str, srcinfo):
        super().__init__(parcellation_id, matrixloader, srcinfo)
        self.paradigm = paradigm


class HcpConnectivityFetcher:

    _DATASET_ID = "0f1ccc4a-9a11-4697-b43f-9c9c8ac543e6"
    # TODO add other parcellations
    _PARCELLATION_NAMES = {
        "294-Julich-Brain": Parcellation.REGISTRY.JULICH_BRAIN_CYTOARCHITECTONIC_MAPS_2_9,
        "096-HarvardOxfordMaxProbThr0": Parcellation.REGISTRY.HARVARDOXFORD_CORT_MAXPROB_THR0,
    }

    def __init__(self, filename_keyword):
        FeatureQuery.__init__(self)
        self._connector = EbrainsPublicDatasetConnector(
            self._DATASET_ID, in_progress=False
        )
        if len(self._connector._files) == 0:
            logger.error(
                f"EBRAINS knowledge graph query for dataset {self._DATASET_ID} "
                "did not return any files. It seems that your EBRAINS authentication "
                "does not provide sufficient privileges to access the connectivity data."
                )
        self._keyword = filename_keyword

    @property
    def doi(self):
        return self._connector.doi

    @property
    def srcinfo(self):
        return {
            "name": self._connector.name,
            "dataset_id": self._DATASET_ID, 
            "description": self._connector.description,
            "citation": self._connector.citation,
            "authors": self._connector.authors,
            "cohort": "HCP"}

    def get_matrixloaders(self, parcellation: Parcellation):
        """Return functions for loading the connectivity matrices
        matching the given parcellation as pandas DataFrames."""

        loaders = []
        for name, parc in self._PARCELLATION_NAMES.items():

            if parc != parcellation:
                continue

            try:
                zipfile = self._connector.get(f"{name}.zip")
            except RuntimeError as e:
                logger.error(str(e))
                continue

            # extract index - regionname mapping
            with zipfile.open(f"{name}/0ImageProcessing/Link.txt") as f:
                lines = [
                    line.decode().strip().split(" ", maxsplit=1)
                    for line in f.readlines()
                ]
                with QUIET:
                    indexmap = {
                        int(line[0]): parc.decode_region(line[1])
                        for line in lines
                        if len(line) == 2 and line[0].isnumeric()
                    }

            # create a dict of the csv files by subject id
            csvfiles = [
                f
                for f in zipfile.namelist()
                if f.endswith(".csv") and (self._keyword in f)
            ]

            for fn in csvfiles:
                subject_id = fn.split("/")[-2]
                # define the lazy loader function
                loaders.append(
                    (
                        fn,
                        subject_id,
                        lambda z=f"{name}.zip", c=fn, m=indexmap, p=parc: self._load_matrix(
                            z, c, m, p
                        ),
                    )
                )

        return loaders

    def _load_matrix(self, zip_filename, csv_filename, indexmap, parcellation):
        """Extract connectivity matrix from a csv file inside a zip file
        known to the dataset connector, and return it as a dataframe with
        region objects as row and column indices.
        """
        logger.debug(f"Loading {csv_filename} from {zip_filename}.")
        zipfile = self._connector.get(zip_filename)
        try:
            matrix = pd.read_csv(
                zipfile.open(csv_filename),
                delimiter=r"\s+|,|;", engine='python',
                header=None,
                index_col=False,
            )
            l = matrix.shape[0]
            if matrix.shape[1] != l:
                raise RuntimeError(
                    f"Non-quadratic connectivity matrix {l}x{matrix.shape[1]} "
                    f"from {csv_filename} in {zip_filename}"
                )
            if len(indexmap) == l:
                remapper = {label - min(indexmap.keys()): region for label, region in indexmap.items()}
                matrix = matrix.rename(index=remapper).rename(columns=remapper)
            else:
                labels = {r.index.label for r in parcellation.regiontree} - {None} 
                if max(labels) - min(labels) + 1 == l:
                    indexmap = {
                        r.index.label - min(labels): r for r in parcellation.regiontree
                        if r.index.label is not None
                    }
                    matrix = matrix.rename(index=indexmap).rename(columns=indexmap)
                else:
                    logger.warn("Could not decode connectivity matrix regions.")

        except pd.errors.ParserError:
            logger.error(
                f"Could not parse connectivity matrix from file {csv_filename} in {zip_filename}."
            )
        return matrix


    @property
    def parcellations(self):
        return list(self._PARCELLATION_NAMES.values())


class HcpStreamlineCountQuery(HcpConnectivityFetcher, FeatureQuery):

    _FEATURETYPE = StreamlineCounts

    def __init__(self, **kwargs):
        FeatureQuery.__init__(self)
        HcpConnectivityFetcher.__init__(self, filename_keyword="Counts")
        for parc in self.parcellations:
            for filename, subject_id, loader in self.get_matrixloaders(parc):
                srcinfo = self.srcinfo
                srcinfo["subject"] = subject_id
                srcinfo["filename"] = filename
                self.register(self._FEATURETYPE(parc, loader, srcinfo))


class HcpStreamlineLengthQuery(HcpConnectivityFetcher, FeatureQuery):

    _FEATURETYPE = StreamlineLengths

    def __init__(self, **kwargs):
        FeatureQuery.__init__(self)
        HcpConnectivityFetcher.__init__(self, filename_keyword="Lengths")
        for parc in self.parcellations:
            for filename, subject_id, loader in self.get_matrixloaders(parc):
                srcinfo = self.srcinfo
                srcinfo["subject"] = subject_id
                srcinfo["filename"] = filename
                self.register(self._FEATURETYPE(parc, loader, srcinfo))


class HcpRestingStateQuery(HcpConnectivityFetcher, FeatureQuery):

    _FEATURETYPE = FunctionalConnectivity

    def __init__(self, **kwargs):
        FeatureQuery.__init__(self)
        HcpConnectivityFetcher.__init__(self, filename_keyword="EmpCorrFC")
        for parc in self.parcellations:
            for filename, subject_id, loader in self.get_matrixloaders(parc):
                paradigm = "Resting state ({})".format(
                    filename.split("/")[-1].replace(".csv", "")
                )
                srcinfo = self.srcinfo
                srcinfo["subject"] = subject_id
                srcinfo["filename"] = filename
                self.register(self._FEATURETYPE(parc, loader, paradigm, srcinfo))


class PrereleasedConnectivityFetcher(GitlabConnector):

    _COHORTS = ["1000brains", "hcp", "enki"]

    def __init__(self, keywords: List[str]):
        """Gitlab connector for accessing some pre-released connectivity matrices."""
        GitlabConnector.__init__(self, "https://jugit.fz-juelich.de", 3009, "develop")
        self._keywords = keywords

    def get_data(self):
        results = []
        for filename, jsonloader in self.get_loaders("connectivity", ".json"):
            if not any(kw.lower() in filename.lower() for kw in self._keywords):
                continue
            parc_id = jsonloader.data["parcellation id"]
            matrixloader = lambda l=jsonloader: self._matrixloader(l)
            srcinfo = {k: v for k, v in jsonloader.data.items() if k != "data"}
            srcinfo["filename"] = filename
            matched_cohorts = [
                c for c in self._COHORTS if filename.lower().find(c.lower()) >= 0
            ]
            if len(matched_cohorts) > 0:
                srcinfo["cohort"] = matched_cohorts[0].upper()
                srcinfo["subject"] = "average"
            results.append((parc_id, matrixloader, srcinfo))
        return results

    def _matrixloader(self, jsonloader):
        """
        Load connectivity matrix from json object
        """
        data = jsonloader.data
        assert "data" in data
        col_names = data["data"]["field names"]
        row_names = list(data["data"]["profiles"].keys())
        assert col_names == row_names, f"{data['name']} assertion error: expected col_names == row_names"
        matrix = pd.DataFrame(
            data=[data["data"]["profiles"][r] for r in col_names],
            columns=col_names,
            index=col_names,
        )
        return matrix


class PrereleasedStreamlineCountQuery(FeatureQuery, PrereleasedConnectivityFetcher):

    _FEATURETYPE = StreamlineCounts

    def __init__(self, **kwargs):
        FeatureQuery.__init__(self)
        PrereleasedConnectivityFetcher.__init__(self, ["count", "1000brains"])
        for parc_id, matrixloader, srcinfo in self.get_data():
            self.register(self._FEATURETYPE(parc_id, matrixloader, srcinfo))


class PrereleasedStreamlineLengthQuery(FeatureQuery, PrereleasedConnectivityFetcher):

    _FEATURETYPE = StreamlineLengths

    def __init__(self, **kwargs):
        FeatureQuery.__init__(self)
        PrereleasedConnectivityFetcher.__init__(self, ["length"])
        for parc_id, matrixloader, srcinfo in self.get_data():
            self.register(self._FEATURETYPE(parc_id, matrixloader, srcinfo))


class PrereleasedRestingStateQuery(FeatureQuery, PrereleasedConnectivityFetcher):

    _FEATURETYPE = FunctionalConnectivity

    def __init__(self, **kwargs):
        FeatureQuery.__init__(self)
        keyword = "rest_fix"
        PrereleasedConnectivityFetcher.__init__(self, [keyword])
        for parc_id, matrixloader, srcinfo in self.get_data():
            filename = srcinfo["filename"]
            paradigm = filename[filename.lower().find(keyword.lower()):]
            self.register(self._FEATURETYPE(parc_id, matrixloader, paradigm, srcinfo))



# generated by datamodel-codegen:
#   filename:  https://raw.githubusercontent.com/HumanBrainProject/openMINDS/3fa86f956b407b2debf47c2e1b6314e37579c707/v3/SANDS/v3/miscellaneous/quantitativeRelationAssessment.schema.json

from typing import Any, Dict, List, Optional, Union

from pydantic import Field
from siibra.openminds.base import SiibraBaseModel


class QuantitativeOverlapItem(SiibraBaseModel):
    type_of_uncertainty: Optional[Any] = Field(
        None,
        alias='typeOfUncertainty',
        description='Distinct technique used to quantify the uncertainty of a measurement.',
        title='typeOfUncertainty',
    )
    uncertainty: Optional[List[float]] = Field(
        None,
        description='Quantitative value range defining the uncertainty of a measurement.',
        max_items=2,
        min_items=2,
        title='uncertainty',
    )
    unit: Optional[Any] = Field(
        None,
        description='Determinate quantity adopted as a standard of measurement.',
        title='unit',
    )
    value: float = Field(..., description='Entry for a property.', title='value')


class QuantitativeOverlapItem1(SiibraBaseModel):
    max_value: float = Field(
        ...,
        alias='maxValue',
        description='Greatest quantity attained or allowed.',
        title='maxValue',
    )
    max_value_unit: Optional[Any] = Field(
        None, alias='maxValueUnit', title='maxValueUnit'
    )
    min_value: float = Field(
        ...,
        alias='minValue',
        description='Smallest quantity attained or allowed.',
        title='minValue',
    )
    min_value_unit: Optional[Any] = Field(
        None, alias='minValueUnit', title='minValueUnit'
    )


class Model(SiibraBaseModel):
    id: str = Field(..., alias='@id', description='Metadata node identifier.')
    type: str = Field(..., alias='@type')
    criteria: Optional[Dict[str, Any]] = Field(
        None,
        alias='criteria',
        description='Aspects or standards on which a judgement or decision is based.',
        title='criteria',
    )
    in_relation_to: Dict[str, Any] = Field(
        ...,
        alias='inRelationTo',
        description='Reference to a related element.',
        title='inRelationTo',
    )
    quantitative_overlap: Union[
        'QuantitativeOverlapItem', 'QuantitativeOverlapItem1'
    ] = Field(..., alias='quantitativeOverlap')



# generated by datamodel-codegen:
#   filename:  https://raw.githubusercontent.com/HumanBrainProject/openMINDS/3fa86f956b407b2debf47c2e1b6314e37579c707/v3/core/v4/research/subjectGroup.schema.json

from typing import Any, List, Optional

from pydantic import Field
from siibra.openminds.base import SiibraBaseModel


class Model(SiibraBaseModel):
    id: str = Field(..., alias='@id', description='Metadata node identifier.')
    type: str = Field(..., alias='@type')
    additional_remarks: Optional[str] = Field(
        None,
        alias='additionalRemarks',
        description='Mention of what deserves additional attention or notice.',
        title='additionalRemarks',
    )
    biological_sex: Optional[List[Any]] = Field(
        None,
        alias='biologicalSex',
        description='Differentiation of individuals of most species (animals and plants) based on the type of gametes they produce.',
        min_items=1,
        title='biologicalSex',
    )
    internal_identifier: Optional[str] = Field(
        None,
        alias='internalIdentifier',
        description='Term or code that identifies someone or something within a particular product.',
        title='internalIdentifier',
    )
    lookup_label: Optional[str] = Field(
        None,
        alias='lookupLabel',
        title='lookupLabel',
    )
    quantity: Optional[int] = Field(
        None,
        alias='quantity',
        description='Total amount or number of things or beings.',
        title='quantity',
    )
    species: List[Any] = Field(
        ...,
        alias='species',
        description='Category of biological classification comprising related organisms or populations potentially capable of interbreeding, and being designated by a binomial that consists of the name of a genus followed by a Latin or latinized uncapitalized noun or adjective.',
        min_items=1,
        title='species',
    )
    studied_state: List[Any] = Field(
        ...,
        alias='studiedState',
        description='Reference to a point in time at which something or someone was studied in a particular mode or condition.',
        min_items=1,
        title='studiedState',
    )



# generated by datamodel-codegen:
#   filename:  https://raw.githubusercontent.com/HumanBrainProject/openMINDS/3fa86f956b407b2debf47c2e1b6314e37579c707/v3/core/v4/products/modelVersion.schema.json

from datetime import date
from typing import Any, Dict, List, Optional

from pydantic import Field, constr
from siibra.openminds.base import SiibraBaseModel


class Copyright(SiibraBaseModel):
    holder: List = Field(
        ...,
        description='Legal person in possession of something.',
        min_items=1,
        title='holder',
    )
    year: constr(regex=r'([0-9]{4})') = Field(
        ...,
        description='Cycle in the Gregorian calendar specified by a number and comprised of 365 or 366 days divided into 12 months beginning with January and ending with December.',
        title='year',
    )


class OtherContribution(SiibraBaseModel):
    contribution_type: List = Field(
        ...,
        alias='contributionType',
        description='Distinct class of what was given or supplied as a part or share.',
        min_items=1,
        title='contributionType',
    )
    contributor: Any = Field(
        ...,
        description='Legal person that gave or supplied something as a part or share.',
        title='contributor',
    )


class Model(SiibraBaseModel):
    id: str = Field(..., alias='@id', description='Metadata node identifier.')
    type: str = Field(..., alias='@type')
    accessibility: Dict[str, Any] = Field(
        ...,
        alias='accessibility',
        description='Level to which something is accessible to someone or something.',
        title='accessibility',
    )
    copyright: Optional['Copyright'] = Field(
        None,
        alias='copyright',
        description='Structured information on the copyright.',
    )
    custodian: Optional[List[Any]] = Field(
        None,
        alias='custodian',
        description="The 'custodian' is a legal person who is responsible for the content and quality of the data, metadata, and/or code of a research product.",
        min_items=1,
        title='custodian',
    )
    description: Optional[constr(max_length=2000)] = Field(
        None,
        alias='description',
        description='Longer statement or account giving the characteristics of someone or something.',
        title='description',
    )
    developer: Optional[List[Any]] = Field(
        None,
        alias='developer',
        description='Legal person that creates or improves products or services (e.g., software, applications, etc.).',
        min_items=1,
        title='developer',
    )
    digital_identifier: Optional[Dict[str, Any]] = Field(
        None,
        alias='digitalIdentifier',
        description='Digital handle to identify objects or legal persons.',
        title='digitalIdentifier',
    )
    format: Dict[str, Any] = Field(
        ...,
        alias='format',
        description='Method of digitally organizing and structuring data or information.',
        title='format',
    )
    full_documentation: Dict[str, Any] = Field(
        ...,
        alias='fullDocumentation',
        description='Non-abridged instructions, comments, and information for using a particular product.',
        title='fullDocumentation',
    )
    full_name: Optional[str] = Field(
        None,
        alias='fullName',
        description='Whole, non-abbreviated name of something or somebody.',
        title='fullName',
    )
    funding: Optional[List[Any]] = Field(
        None,
        alias='funding',
        description='Money provided by a legal person for a particular purpose.',
        min_items=1,
        title='funding',
    )
    homepage: Optional[Dict[str, Any]] = Field(
        None,
        alias='homepage',
        description='Main website of something or someone.',
        title='homepage',
    )
    how_to_cite: Optional[str] = Field(
        None,
        alias='howToCite',
        description='Preferred format for citing a particular object or legal person.',
        title='howToCite',
    )
    input_data: Optional[List[Any]] = Field(
        None,
        alias='inputData',
        description='Data that is put into a process or machine.',
        min_items=1,
        title='inputData',
    )
    is_alternative_version_of: Optional[List[Any]] = Field(
        None,
        alias='isAlternativeVersionOf',
        description='Reference to an original form where the essence was preserved, but presented in an alternative form.',
        min_items=1,
        title='isAlternativeVersionOf',
    )
    is_new_version_of: Optional[Dict[str, Any]] = Field(
        None,
        alias='isNewVersionOf',
        description='Reference to a previous (potentially outdated) particular form of something.',
        title='isNewVersionOf',
    )
    keyword: Optional[List[Any]] = Field(
        None,
        alias='keyword',
        description='Significant word or concept that are representative of something or someone.',
        max_items=5,
        min_items=1,
        title='keyword',
    )
    license: List[Any] = Field(
        ...,
        alias='license',
        description='Grant by a party to another party as an element of an agreement between those parties that permits to do, use, or own something.',
        min_items=1,
        title='license',
    )
    other_contribution: Optional['OtherContribution'] = Field(
        None,
        alias='otherContribution',
        description='Structured information on the contribution made to a research product.',
    )
    output_data: Optional[List[Any]] = Field(
        None,
        alias='outputData',
        description='Data that comes out of, is delivered or produced by a process or machine.',
        min_items=1,
        title='outputData',
    )
    related_publication: Optional[List[Any]] = Field(
        None,
        alias='relatedPublication',
        description='Reference to something that was made available for the general public to see or buy.',
        min_items=1,
        title='relatedPublication',
    )
    release_date: date = Field(
        ...,
        alias='releaseDate',
        description='Fixed date on which a product is due to become or was made available for the general public to see or buy',
        title='releaseDate',
    )
    repository: Optional[Dict[str, Any]] = Field(
        None,
        alias='repository',
        description='Place, room, or container where something is deposited or stored.',
        title='repository',
    )
    short_name: constr(max_length=30) = Field(
        ...,
        alias='shortName',
        description='Shortened or fully abbreviated name of something or somebody.',
        title='shortName',
    )
    support_channel: Optional[List[str]] = Field(
        None,
        alias='supportChannel',
        description='Way of communication used to interact with users or customers.',
        min_items=1,
        title='supportChannel',
    )
    version_identifier: str = Field(
        ...,
        alias='versionIdentifier',
        description='Term or code used to identify the version of something.',
        title='versionIdentifier',
    )
    version_innovation: str = Field(
        ...,
        alias='versionInnovation',
        description='Documentation on what changed in comparison to a previously published form of something.',
        title='versionInnovation',
    )



# generated by datamodel-codegen:
#   filename:  https://raw.githubusercontent.com/HumanBrainProject/openMINDS/3fa86f956b407b2debf47c2e1b6314e37579c707/v3/SANDS/v3/atlas/atlasAnnotation.schema.json

from typing import Any, Dict, List, Optional

from pydantic import Field, constr
from siibra.openminds.base import SiibraBaseModel


class Coordinates(SiibraBaseModel):
    type_of_uncertainty: Optional[Any] = Field(
        None,
        alias='typeOfUncertainty',
        description='Distinct technique used to quantify the uncertainty of a measurement.',
        title='typeOfUncertainty',
    )
    uncertainty: Optional[List[float]] = Field(
        None,
        description='Quantitative value range defining the uncertainty of a measurement.',
        max_items=2,
        min_items=2,
        title='uncertainty',
    )
    unit: Optional[Any] = Field(
        None,
        description='Determinate quantity adopted as a standard of measurement.',
        title='unit',
    )
    value: float = Field(..., description='Entry for a property.', title='value')


class BestViewPoint(SiibraBaseModel):
    coordinate_space: Any = Field(
        ...,
        alias='coordinateSpace',
        description='Two or three dimensional geometric setting.',
        title='coordinateSpace',
    )
    coordinates: 'Coordinates' = Field(
        ..., description='Structured information on a quantitative value.'
    )


class Model(SiibraBaseModel):
    id: str = Field(..., alias='@id', description='Metadata node identifier.')
    type: str = Field(..., alias='@type')
    best_view_point: Optional['BestViewPoint'] = Field(
        None,
        alias='bestViewPoint',
        description='Structured information on a coordinate point.',
    )
    criteria: Optional[Dict[str, Any]] = Field(
        None,
        alias='criteria',
        description='Aspects or standards on which a judgement or decision is based.',
        title='criteria',
    )
    criteria_quality_type: Dict[str, Any] = Field(
        ...,
        alias='criteriaQualityType',
        description='Distinct class that defines how the judgement or decision was made for a particular criteria.',
        title='criteriaQualityType',
    )
    display_color: Optional[constr(regex=r'^#[0-9A-Fa-f]{6}$')] = Field(
        None,
        alias='displayColor',
        description='Preferred coloring.',
        title='displayColor',
    )
    inspired_by: Optional[List[Any]] = Field(
        None,
        alias='inspiredBy',
        description='Reference to an inspiring element.',
        min_items=1,
        title='inspiredBy',
    )
    internal_identifier: str = Field(
        ...,
        alias='internalIdentifier',
        description='Term or code that identifies someone or something within a particular product.',
        title='internalIdentifier',
    )
    laterality: Optional[List[Any]] = Field(
        None,
        alias='laterality',
        description='Differentiation between a pair of lateral homologous parts of the body.',
        max_items=2,
        min_items=1,
        title='laterality',
    )
    visualized_in: Optional[Dict[str, Any]] = Field(
        None,
        alias='visualizedIn',
        description='Reference to an image in which something is visible.',
        title='visualizedIn',
    )



# Copyright 2018-2021
# Institute of Neuroscience and Medicine (INM-1), Forschungszentrum Jülich GmbH

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .feature import SpatialFeature
from .query import FeatureQuery

from .. import QUIET, logger
from ..volumes.volume import VolumeSrc, VolumeModel, ColorVolumeNotSupported
from ..core.space import BoundingBoxModel, Space, BoundingBox
from ..core.datasets import EbrainsDataset, DatasetJsonModel
from ..retrieval.repositories import GitlabConnector
from ..core.serializable_concept import JSONSerializable

import numpy as np
from typing import List
import hashlib
from pydantic import Field


class VolumeOfInterest(SpatialFeature, EbrainsDataset, JSONSerializable):
    def __init__(self, dataset_id, location, name):
        SpatialFeature.__init__(self, location)
        EbrainsDataset.__init__(self, dataset_id, name)
        self.volumes = []

    @classmethod
    def _from_json(cls, definition):
        if definition["@type"] == "minds/core/dataset/v1.0.0":
            space = Space.REGISTRY[definition["space id"]]
            vsrcs = []
            minpoints = []
            maxpoints = []
            for vsrc_def in definition["volumeSrc"]:
                try:
                    vsrc = VolumeSrc._from_json(vsrc_def)
                    vsrc.space = space
                    with QUIET:
                        img = vsrc.fetch()
                        D = np.asanyarray(img.dataobj).squeeze()
                        nonzero = np.array(np.where(D > 0))
                        A = img.affine
                    minpoints.append(np.dot(A, np.r_[nonzero.min(1)[:3], 1])[:3])
                    maxpoints.append(np.dot(A, np.r_[nonzero.max(1)[:3], 1])[:3])
                    vsrcs.append(vsrc)
                    
                except ColorVolumeNotSupported:
                    # If multi channel volume exists rather than short circuit, try to use other volumes to determine the ROI
                    # See PLI hippocampus data feature
                    vsrcs.append(vsrc)
                    
            minpoint = np.array(minpoints).min(0)
            maxpoint = np.array(maxpoints).max(0)
            result = cls(
                dataset_id=definition["kgId"],
                name=definition["name"],
                location=BoundingBox(minpoint, maxpoint, space),
            )
            list(map(result.volumes.append, vsrcs))
            return result
        return definition

    @classmethod
    def get_model_type(Cls):
        return "siibra/features/voi"

    @property
    def model_id(self):
        _id = hashlib.md5(super().model_id.encode("utf-8")).hexdigest()
        return f"{VolumeOfInterest.get_model_type()}/{str(_id)}"

    def to_model(self, **kwargs) -> 'VOIDataModel':
        super_model = super().to_model(**kwargs)
        super_model_dict = super_model.dict()
        super_model_dict["@type"] = VolumeOfInterest.get_model_type()
        return VOIDataModel(
            location=self.location.to_model(**kwargs),
            volumes=[vol.to_model(**kwargs) for vol in self.volumes],
            **super_model_dict,
        )


class VOIDataModel(DatasetJsonModel):
    type: str = Field(VolumeOfInterest.get_model_type(), const=True, alias="@type")
    volumes: List[VolumeModel]
    location: BoundingBoxModel


class VolumeOfInterestQuery(FeatureQuery):
    _FEATURETYPE = VolumeOfInterest
    _QUERY = GitlabConnector("https://jugit.fz-juelich.de", 3009, "develop")

    def __init__(self, **kwargs):
        FeatureQuery.__init__(self)
        for _, loader in self._QUERY.get_loaders(folder="vois", suffix=".json"):
            try:
                voi = VolumeOfInterest._from_json(loader.data)  # json.loads(data))
                self.register(voi)
            except Exception as e:
                logger.warn(f"some VOI cannot be loaded: {str(e)}")



# generated by datamodel-codegen:
#   filename:  https://raw.githubusercontent.com/HumanBrainProject/openMINDS/3fa86f956b407b2debf47c2e1b6314e37579c707/v3/core/v4/data/contentType.schema.json

from typing import Any, List, Optional

from pydantic import Field
from siibra.openminds.base import SiibraBaseModel


class Model(SiibraBaseModel):
    id: str = Field(..., alias='@id', description='Metadata node identifier.')
    type: str = Field(..., alias='@type')
    data_type: Optional[List[Any]] = Field(
        None,
        alias='dataType',
        min_items=1,
        title='dataType',
    )
    description: Optional[str] = Field(
        None,
        alias='description',
        description='Longer statement or account giving the characteristics of someone or something.',
        title='description',
    )
    file_extension: Optional[List[str]] = Field(
        None,
        alias='fileExtension',
        description='String of characters attached as suffix to the names of files of a particular format.',
        min_items=1,
        title='fileExtension',
    )
    name: str = Field(
        ...,
        alias='name',
        description='Word or phrase that constitutes the distinctive designation of a being or thing.',
        title='name',
    )
    related_media_type: Optional[str] = Field(
        None,
        alias='relatedMediaType',
        description='Reference to an official two-part identifier for file formats and format contents.',
        title='relatedMediaType',
    )
    specification: Optional[str] = Field(
        None,
        alias='specification',
        description='Detailed and precise presentation of, or proposal for something.',
        title='specification',
    )
    synonym: Optional[List[str]] = Field(
        None,
        alias='synonym',
        description='Words or expressions used in the same language that have the same or nearly the same meaning in some or all senses.',
        min_items=1,
        title='synonym',
    )



# generated by datamodel-codegen:
#   filename:  https://raw.githubusercontent.com/HumanBrainProject/openMINDS/3fa86f956b407b2debf47c2e1b6314e37579c707/v3/SANDS/v3/non-atlas/customCoordinateSpace.schema.json

from typing import Any, Dict, List, Optional

from pydantic import Field
from siibra.openminds.base import SiibraBaseModel


class AxesOrigin(SiibraBaseModel):
    type_of_uncertainty: Optional[Any] = Field(
        None,
        alias='typeOfUncertainty',
        description='Distinct technique used to quantify the uncertainty of a measurement.',
        title='typeOfUncertainty',
    )
    uncertainty: Optional[List[float]] = Field(
        None,
        description='Quantitative value range defining the uncertainty of a measurement.',
        max_items=2,
        min_items=2,
        title='uncertainty',
    )
    unit: Optional[Any] = Field(
        None,
        description='Determinate quantity adopted as a standard of measurement.',
        title='unit',
    )
    value: float = Field(..., description='Entry for a property.', title='value')


class Model(SiibraBaseModel):
    id: str = Field(..., alias='@id', description='Metadata node identifier.')
    type: str = Field(..., alias='@type')
    anatomical_axes_orientation: Dict[str, Any] = Field(
        ...,
        alias='anatomicalAxesOrientation',
        description='Relation between reference planes used in anatomy and mathematics.',
        title='anatomicalAxesOrientation',
    )
    axes_origin: 'AxesOrigin' = Field(
        ...,
        alias='axesOrigin',
        description='Structured information on a quantitative value.',
    )
    default_image: Optional[List[Any]] = Field(
        None,
        alias='defaultImage',
        description='Two or three dimensional image that particluarly represents a specific coordinate space.',
        min_items=1,
        title='defaultImage',
    )
    name: str = Field(
        ...,
        alias='name',
        description='Word or phrase that constitutes the distinctive designation of a being or thing.',
        title='name',
    )
    native_unit: Dict[str, Any] = Field(
        ...,
        alias='nativeUnit',
        description='Determinate quantity used in the original measurement.',
        title='nativeUnit',
    )



# generated by datamodel-codegen:
#   filename:  https://raw.githubusercontent.com/HumanBrainProject/openMINDS/3fa86f956b407b2debf47c2e1b6314e37579c707/v3/core/v4/research/subjectState.schema.json

from typing import Any, Dict, List, Optional, Union

from pydantic import Field
from siibra.openminds.base import SiibraBaseModel


class AgeItem(SiibraBaseModel):
    type_of_uncertainty: Optional[Any] = Field(
        None,
        alias='typeOfUncertainty',
        description='Distinct technique used to quantify the uncertainty of a measurement.',
        title='typeOfUncertainty',
    )
    uncertainty: Optional[List[float]] = Field(
        None,
        description='Quantitative value range defining the uncertainty of a measurement.',
        max_items=2,
        min_items=2,
        title='uncertainty',
    )
    unit: Optional[Any] = Field(
        None,
        description='Determinate quantity adopted as a standard of measurement.',
        title='unit',
    )
    value: float = Field(..., description='Entry for a property.', title='value')


class AgeItem1(SiibraBaseModel):
    max_value: float = Field(
        ...,
        alias='maxValue',
        description='Greatest quantity attained or allowed.',
        title='maxValue',
    )
    max_value_unit: Optional[Any] = Field(
        None, alias='maxValueUnit', title='maxValueUnit'
    )
    min_value: float = Field(
        ...,
        alias='minValue',
        description='Smallest quantity attained or allowed.',
        title='minValue',
    )
    min_value_unit: Optional[Any] = Field(
        None, alias='minValueUnit', title='minValueUnit'
    )


class WeightItem(SiibraBaseModel):
    type_of_uncertainty: Optional[Any] = Field(
        None,
        alias='typeOfUncertainty',
        description='Distinct technique used to quantify the uncertainty of a measurement.',
        title='typeOfUncertainty',
    )
    uncertainty: Optional[List[float]] = Field(
        None,
        description='Quantitative value range defining the uncertainty of a measurement.',
        max_items=2,
        min_items=2,
        title='uncertainty',
    )
    unit: Optional[Any] = Field(
        None,
        description='Determinate quantity adopted as a standard of measurement.',
        title='unit',
    )
    value: float = Field(..., description='Entry for a property.', title='value')


class WeightItem1(SiibraBaseModel):
    max_value: float = Field(
        ...,
        alias='maxValue',
        description='Greatest quantity attained or allowed.',
        title='maxValue',
    )
    max_value_unit: Optional[Any] = Field(
        None, alias='maxValueUnit', title='maxValueUnit'
    )
    min_value: float = Field(
        ...,
        alias='minValue',
        description='Smallest quantity attained or allowed.',
        title='minValue',
    )
    min_value_unit: Optional[Any] = Field(
        None, alias='minValueUnit', title='minValueUnit'
    )


class Model(SiibraBaseModel):
    id: str = Field(..., alias='@id', description='Metadata node identifier.')
    type: str = Field(..., alias='@type')
    additional_remarks: Optional[str] = Field(
        None,
        alias='additionalRemarks',
        description='Mention of what deserves additional attention or notice.',
        title='additionalRemarks',
    )
    age: Optional[Union['AgeItem', 'AgeItem1']] = Field(
        None, alias='age'
    )
    age_category: Dict[str, Any] = Field(
        ...,
        alias='ageCategory',
        description='Distinct life cycle class that is defined by a similar age or age range (developmental stage) within a group of individual beings.',
        title='ageCategory',
    )
    attribute: Optional[List[Any]] = Field(
        None,
        alias='attribute',
        min_items=1,
        title='attribute',
    )
    handedness: Optional[Dict[str, Any]] = Field(
        None,
        alias='handedness',
        description='Degree to which an organism prefers one hand or foot over the other hand or foot during the performance of a task.',
        title='handedness',
    )
    lookup_label: Optional[str] = Field(
        None,
        alias='lookupLabel',
        title='lookupLabel',
    )
    pathology: Optional[List[Any]] = Field(
        None,
        alias='pathology',
        description='Structural and functional deviation from the normal that constitutes a disease or characterizes a particular disease.',
        min_items=1,
        title='pathology',
    )
    weight: Optional[Union['WeightItem', 'WeightItem1']] = Field(
        None, alias='weight'
    )



# generated by datamodel-codegen:
#   filename:  https://raw.githubusercontent.com/HumanBrainProject/openMINDS/3fa86f956b407b2debf47c2e1b6314e37579c707/v3/core/v4/miscellaneous/SWHID.schema.json

from pydantic import Field, constr
from siibra.openminds.base import SiibraBaseModel


class Model(SiibraBaseModel):
    id: str = Field(..., alias='@id', description='Metadata node identifier.')
    type: str = Field(..., alias='@type')
    identifier: constr(
        regex=r'^https://archive.softwareheritage.org/swh:1:(cnt|dir|rel|rev|snp):[0-9a-f]{40}(;(origin|visit|anchor|path|lines)=[^ \t\r\n\f]+)*$'
    ) = Field(
        ...,
        alias='identifier',
        description='Term or code used to identify something or someone.',
        title='identifier',
    )



# generated by datamodel-codegen:
#   filename:  https://raw.githubusercontent.com/HumanBrainProject/openMINDS/3fa86f956b407b2debf47c2e1b6314e37579c707/v3/core/v4/actors/person.schema.json

from typing import Any, Dict, List, Optional

from pydantic import Field
from siibra.openminds.base import SiibraBaseModel


class Affiliation(SiibraBaseModel):
    end_date: Optional[str] = Field(
        None,
        alias='endDate',
        description='Date in the Gregorian calendar at which something terminates in time.',
        title='endDate',
    )
    organization: Any = Field(
        ...,
        description='Legally accountable, administrative and functional structure.',
        title='organization',
    )
    start_date: Optional[str] = Field(
        None,
        alias='startDate',
        description='Date in the Gregorian calendar at which something begins in time',
        title='startDate',
    )


class Model(SiibraBaseModel):
    id: str = Field(..., alias='@id', description='Metadata node identifier.')
    type: str = Field(..., alias='@type')
    affiliation: Optional['Affiliation'] = Field(
        None, alias='affiliation'
    )
    contact_information: Optional[Dict[str, Any]] = Field(
        None,
        alias='contactInformation',
        description='Any available way used to contact a person or business (e.g., address, phone number, email address, etc.).',
        title='contactInformation',
    )
    digital_identifier: Optional[List[Any]] = Field(
        None,
        alias='digitalIdentifier',
        description='Digital handle to identify objects or legal persons.',
        min_items=1,
        title='digitalIdentifier',
    )
    family_name: Optional[str] = Field(
        None,
        alias='familyName',
        description='Name borne in common by members of a family.',
        title='familyName',
    )
    given_name: str = Field(
        ...,
        alias='givenName',
        description='Name given to a person, including all potential middle names, but excluding the family name.',
        title='givenName',
    )



# generated by datamodel-codegen:
#   filename:  https://raw.githubusercontent.com/HumanBrainProject/openMINDS/3fa86f956b407b2debf47c2e1b6314e37579c707/v3/core/v4/actors/contribution.schema.json

from typing import Any, Dict, List

from pydantic import Field
from siibra.openminds.base import SiibraBaseModel


class Model(SiibraBaseModel):
    id: str = Field(..., alias='@id', description='Metadata node identifier.')
    type: str = Field(..., alias='@type')
    contribution_type: List[Any] = Field(
        ...,
        alias='contributionType',
        description='Distinct class of what was given or supplied as a part or share.',
        min_items=1,
        title='contributionType',
    )
    contributor: Dict[str, Any] = Field(
        ...,
        alias='contributor',
        description='Legal person that gave or supplied something as a part or share.',
        title='contributor',
    )



# generated by datamodel-codegen:
#   filename:  https://raw.githubusercontent.com/HumanBrainProject/openMINDS/3fa86f956b407b2debf47c2e1b6314e37579c707/v3/computation/v1/launchConfiguration.schema.json

from typing import Any, Dict, List, Optional

from pydantic import Field
from siibra.openminds.base import SiibraBaseModel


class Model(SiibraBaseModel):
    id: str = Field(..., alias='@id', description='Metadata node identifier.')
    type: str = Field(..., alias='@type')
    arguments: Optional[List[str]] = Field(
        None,
        alias='arguments',
        min_items=1,
        title='arguments',
    )
    description: Optional[str] = Field(
        None,
        alias='description',
        description='Longer statement or account giving the characteristics of someone or something.',
        title='description',
    )
    environment_variables: Optional[Dict[str, Any]] = Field(
        None,
        alias='environmentVariables',
        title='environmentVariables',
    )
    executable: str = Field(
        ..., alias='executable', title='executable'
    )
    name: Optional[str] = Field(
        None,
        alias='name',
        description='Word or phrase that constitutes the distinctive designation of a being or thing.',
        title='name',
    )
